#!/bin/sh
# the next line restarts using wish \
exec wish "$0" "$@"

#############################################################################
#
# Trebuchet Tk
# copyright 1997-2003 by Fuzzball Software
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
#############################################################################

package require opt
catch { package require Img }

#################################
# GLOBAL VARIABLES
#
global treb_revision; set treb_revision 1052
global treb_version; set treb_version "1.[format %03d [expr {$treb_revision - 1000}]]"
global treb_name; set treb_name "Trebuchet Tk"
global wordchars; set wordchars {-A-Za-z0-9_'\\.}
global wordchar; set wordchar "\[$wordchars\]"
global nonwordchar; set nonwordchar "\[^$wordchars\]"

set url_end {[^() ]*[a-z0-9_/]}
set domain_chars {[a-z0-9_.-]}
set domain_start "\[a-z0-9_-\]$domain_chars*\[a-z0-9_-\]"
set ftp_regexp "ftp://$url_end|ftp\\.$url_end"
set url_domains "$domain_start\\.\[a-z\]\[a-z\]\[\[:>:\]\]"
append url_domains "|$domain_start\\.\[a-z\]\[a-z\]/$url_end"
foreach tld [list com net org edu gov mil int biz info name museum coop aero] {
    append url_domains "|$domain_start\\.$tld\[\[:>:\]\]"
    append url_domains "|$domain_start\\.$tld/$url_end"
}
set web_regexp "http://$url_end|www\\.$url_end|w3\\.$url_end|web\\.$url_end|$url_domains"
set web2_regexp "https://$url_end"
set rtsp_regexp "rtsp://$url_end"
set mail_regexp "mailto:\[a-z0-9_+.:-\]*@$domain_chars*\[a-z0-9\]"
set mail_regexp2 "\[a-z0-9\]\[a-z0-9_+.:-\]*@$domain_chars*\[a-z0-9\]"
global url_regexp
set url_regexp "$ftp_regexp|$web_regexp|$web2_regexp|$mail_regexp|$mail_regexp2|$rtsp_regexp"

global errors_nonfatal; set errors_nonfatal 0
global cmdhist; set cmdhist {}
global cmdhistnum; set cmdhistnum -1
global defpri; set defpri {10}
global mpid; set mpid 1
global dirty_preferences; set dirty_preferences 0
global sok; set sok {}
global styleslist; set styleslist {}
global widget; 
    set widget(disp) {.mw.con.top.con.disp}
    set widget(backdrop) {.mw.con.top.con.disp.backdrop}
    set widget(inbuf) {.mw.con.bot.con.in.inbuf}
    set widget(statbar) {.mw.con.top.con.status}
    set widget(loglight) {.mw.statusbar.log}
    set widget(worldsbar) {.mw.con.top.con.worlds}
global treb_colors

#################################

proc find_windows_prefsfile {} {
    global treb_prefs_dir treb_root_dir treb_save_file

    set treb_save_file {}
    package require registry 1.0
    set key "HKEY_CURRENT_USER\\Software\\Fuzzball Software\\Trebuchet Tk\\1.0"
    if {![catch {registry get $key "prefsfile"} file]} {
        if {[file isfile $file]} {
            if {[file readable $file]} {
                set treb_save_file $file
                return ""
            } else {
                tk_messageBox -type "ok" -title "Preferences File Unreadable" \
                    -message "The file '$file' isn't readable.  Please specify a different file, or fix the permissions."
            }
        }
    } else {
        # No registry entry.  Probably first time run on this machine.
        return ""
    }

    # Prefs file was moved or is now unreadable.
    set filetypes {
        {{Trebuchet Preferences Files}       {.trc}    TEXT}
    }
    set initdir $treb_prefs_dir
    set initfile "trebprefs.trc"
    if {$file != "" && [file isfile $file]} {
        set initdir [file dirname $file]
        set initfile [file tail $file]
    } elseif {[file isfile [file join $treb_root_dir "trebpref.trc"]]} {
        set initdir $treb_root_dir
    }
    while (1) {
        set dofind [tk_messageBox -title {Trebuchet Preferences File} \
                        -message "Unable to open Trebuchet's preferences file.  Would you like to specify its location?" \
                        -type yesno -icon warning -default "no"]

        if {$dofind == "no"} {
            catch {registry delete $key "prefsfile"}
            set treb_save_file {}
            return ""
        }

        set treb_save_file [tk_getOpenFile -defaultextension .trc \
                                -initialdir $initdir \
                                -initialfile "trebpref.trc" \
                                -title {Specify Trebuchet Preferences File} \
                                -filetypes $filetypes]

        if {$treb_save_file != ""} {
            set dir [file dirname $treb_save_file]
            if {![file isdirectory $dir]} {
                tk_messageBox -type "ok" -title "Preferences File Unwritable" \
                    -message "The directory '$dir' does not exist."
            } elseif {![file writable $dir]} {
                tk_messageBox -type "ok" -title "Preferences File Unwritable" \
                    -message "You do not have permission to write to the directory '$dir'.  You will need to save your prefs to a different directory or, fix the directory permissions."
            } elseif {![file readable $treb_save_file]} {
                tk_messageBox -type "ok" -title "Preferences File Unreadable" \
                    -message "The file '$file' isn't readable.  Please specify a different file, or fix the permissions."
            } elseif {![file writable $treb_save_file]} {
                tk_messageBox -type "ok" -title "Preferences File Unwritable" \
                    -message "You do not have permission to write to the file '$file'.  You will need to save your prefs to a different file, later, or fix the file permissions."
            } else {
                registry set $key "prefsfile" $treb_save_file
                return ""
            }
        }
    }
}


proc /clock_clicks {} {
    if {![catch {clock clicks -milliseconds} millis]} {
        return $millis
    } else {
        return [clock clicks]
    }
}

proc calculate_clicks_per_second {} {
    global clicks_per_second

    if {![catch {clock clicks -milliseconds}]} {
        set clicks_per_second 1000
    } else {
        set now [clock seconds]
        while {$now == [clock seconds]} {
            set sclick [clock clicks]
        }
        set now [clock seconds]
        while {$now == [clock seconds]} {
            set eclick [clock clicks]
        }
        set clicks_per_second [expr {$eclick - $sclick}]
    }
}

proc init {argc argv} {
    global tcl_platform tcl_version
    global treb_colors treb_save_file
    global treb_root_dir treb_lib_dir
    global treb_docs_dir treb_document_dir
    global treb_tool_dir treb_temp_dir
    global treb_cacerts_dir treb_prefs_dir
    global argv0 env treb_fonts

    set tclpatchlevel [info patchlevel]
    set pattern {^([0-9][0-9]*)[^0-9]([0-9][0-9]*)([^0-9])([0-9][0-9]*)}
    regexp $pattern $tclpatchlevel dummy vermajor verminor separ patchlevel
    if {$separ != "."} {
        set patchlevel 0
    }
    set tcl_platform(vermajor) $vermajor
    set tcl_platform(verminor) $verminor
    set tcl_platform(patchlevel) $patchlevel

    if {[info exists env(TREB_ROOT_DIR)]} {
        set treb_root_dir $env(TREB_ROOT_DIR)
    } elseif {$argv0 != {}} {
        set myexec $argv0

        # Make sure we find the REAL Trebuchet.tcl, and not a softlink.
        while {![catch {file readlink $myexec}]} {
            set myexec [file join \
            [file dirname $myexec] \
            [file readlink $myexec]]
        }

        # The root trebuchet dir is the dir that Trebuchet.tcl is in.
        set treb_root_dir [file dirname $myexec]
        if {$tcl_platform(os) == "Darwin"} {
            set treb_root_dir [file join $treb_root_dir .. .. .. trebuchet]
        }
    } else {
        tk_messageBox -default ok -type ok -icon error \
            -title {Trebuchet error} -message "Trebuchet was unable to find its libraries.\nYou may need to set your TREB_ROOT_DIR environment variable."
        exit
    }

    if {$tcl_platform(platform) == "unix" && $tcl_platform(os) == "Darwin"} {
        # The "correct" place for document files on the mac (OS X)
        # is in ~/Library/Preferences folder
        set treb_document_dir [file join $env(HOME) Documents]
    } elseif {$tcl_platform(platform) == "macintosh"} {
        # The "correct" place for document files on the mac (OS 9)
        # is in the application's directory.
        set treb_document_dir $env(HOME)
    } elseif {[info exists env(HOME)]} {
        # For Windows and Unix, use the user's home dir.
        set treb_document_dir $env(HOME)
        if {![file exists $treb_document_dir]} {
            set treb_document_dir $treb_root_dir
        }
    } else {
        set treb_document_dir $treb_root_dir
    }

    if {$tcl_platform(platform) == "unix" && $tcl_platform(os) == "Darwin"} {
        # The "correct" place for preferences files on the mac (OS X)
        # is in ~/Library/Preferences folder
        set treb_prefs_dir [file join $env(HOME) Library Preferences]
    } elseif {$tcl_platform(platform) == "macintosh"} {
        # The "correct" place for preferences files on the mac (OS 9)
        # is in the preferences folder
        set treb_prefs_dir $env(PREF_FOLDER)
    } elseif {[info exists env(HOME)]} {
        set treb_prefs_dir $env(HOME)
        if {$tcl_platform(platform) == "windows"} {
            if {![file exists $treb_prefs_dir]} {
                set treb_prefs_dir $treb_root_dir
            }
        }
    } else {
        set treb_prefs_dir $treb_root_dir
    }

    set treb_lib_dir [file join $treb_root_dir lib]
    set treb_docs_dir [file join $treb_root_dir docs]
    set treb_tool_dir [file join $treb_root_dir pkgs]

# Under OS X 10.2.6 on my powerbook, I get the following weirdness:
#   Windows 98 Tk (8.3.3) (under VPC6) thinks my monitor is 96dpi.
#   Linux Tk (8.3.3) (under VPC6) thinks my monitor is 75dpi by default.
#   OS X Aqua Tk (8.4.4) thinks my monitor is 75dpi.
#   My monitor is actually 100dpi.
# Grrr.

    label .fontcheck
    array set tmp_font [font actual [lrange [.fontcheck cget -font] 0 1]]
    set sansfont [list $tmp_font(-family) $tmp_font(-size)]
    set sansh [font metrics $sansfont -linespace]
    if {[font metrics $sansfont -fixed]} {
        array set tmp_font [font actual [list Helvetica -$sansh]]
        set sansfont [list $tmp_font(-family) $tmp_font(-size)]
    }
    set treb_fonts(sansserif) $sansfont
    if {$tcl_platform(platform) != "unix" || $tcl_platform(os) == "Darwin"} {
        array set tmp_font [font actual [list $tmp_font(-family) [expr {int(0.8*$tmp_font(-size))}]]]
        set bbarfont [list $tmp_font(-family) $tmp_font(-size)]
        set treb_fonts(bbar) $bbarfont
    } else {
        set treb_fonts(bbar) $sansfont
    }
    destroy .fontcheck
    unset tmp_font

    text .fontcheck
    array set tmp_font [font actual [lrange [.fontcheck cget -font] 0 1]]
    set fixedfont [list $tmp_font(-family) $tmp_font(-size)]
    set fixedh [font metrics $fixedfont -linespace]
    if {![font metrics $fixedfont -fixed]} {
        array set tmp_font [font actual [list Courier -$fixedh]]
        set fixedfont [list $tmp_font(-family) $tmp_font(-size)]
    }
    set treb_fonts(fixed) $fixedfont
    destroy .fontcheck
    unset tmp_font

    array set tmp_font [font actual [list Times -$sansh]]
    set treb_fonts(serif) [list $tmp_font(-family) $tmp_font(-size)]
    unset tmp_font

    if {[lsearch -exact $argv "--tkshell"] != -1} {
        return
    }

    source [file join $treb_lib_dir webview.tcl]

    if {$vermajor != 8 || ($verminor == 0 && $patchlevel <= 2)} {
        set mesg "Trebuchet Tk requires TCL/Tk version 8.3 or later to run."
        append mesg "\nPlease fetch and install the latest TCL/Tk 8 interpreter from:"
        set tclurl "http://www.belfry.com/fuzzball/trebuchet/faq.html#WhereGetIt"
        label .icon -bitmap error -foreground red
        label .text -text $mesg -anchor sw -justify left -font $treb_fonts(sansserif)
        label .url -text $tclurl -anchor nw -justify left -foreground blue -font [concat $treb_fonts(sansserif) underline] -cursor hand2
        button .btn -text Okay -command exit
        bind .url <ButtonPress-1> "
                .url config -foreground red
                /web_view [list $tclurl]
                after 250 .url config -foreground blue
            "
        pack .btn -side bottom -anchor s -padx 10 -pady 10
        pack .icon -side left -anchor w -padx 10 -pady 10
        pack .text -side top -anchor w -expand 1 -fill x -padx 10 -pady 5
        pack .url -side top -anchor nw -expand 1 -fill both -padx 10 -pady 0
        wm title . "TCL Version Error"
        bell
        vwait tcl_version
        exit
    }
    wm withdraw .
    global treb_web_cache_dir
    global treb_web_cache_map
    if {$tcl_platform(platform) == "unix"} {
        set treb_web_cache_dir [file join $treb_prefs_dir .trebtk-web-cache]
    } else {
        set treb_web_cache_dir [file join $treb_prefs_dir trebcache]
    }
    if {![file exists $treb_web_cache_dir]} {
        file mkdir $treb_web_cache_dir
    }
    global treb_web_cache_index
    set treb_web_cache_index [file join $treb_web_cache_dir index]
    if {[file exists $treb_web_cache_index]} {
        set f [open $treb_web_cache_index "r"]
        set map_version [gets $f]
        switch -exact $map_version {
            "1" { array set treb_web_cache_map [read $f] }
        }
        close $f
    }

    switch -glob -- $tcl_platform(os) {
        Win* {
            set treb_colors(buttonface) systembuttonface
            set treb_colors(window)     systemwindow
            set treb_colors(windowtext) systemwindowtext
            find_windows_prefsfile

            if {[info exists env(TEMP)]} {
                set treb_temp_dir $env(TEMP)
            } else {
                set treb_temp_dir $treb_root_dir
            }
            if {[catch {event add <<ContextMenu>> <Button-3> <Key-App>}]} {
                event add <<ContextMenu>> <Button-3>
            }
            bind Text <Control-Key-comma> { text_line_shift -4 %W }
            bind Text <Control-Key-period> { text_line_shift 4 %W }
            bind Text <Control-Key-less> { text_line_shift -4 %W }
            bind Text <Control-Key-greater> { text_line_shift 4 %W }
            bind Text <ButtonRelease-1> {+
                if {[/prefs:get copy_on_select]} {
                    event generate %W <<Copy>>
                }
            }
            bind Text <ButtonPress-2> {+
                if {[/prefs:get button2_paste]} {
                    event generate %W <<PasteAt>> -x %x -y %y
                    break
                }
            }
            event add <<Cut>>   <Shift-Key-Delete> 
            event add <<Copy>>  <Control-Key-Insert> 
            event add <<Paste>> <Shift-Key-Insert> 
        }
        Darwin* -
        Mac* {
            # tk scaling 1.4
            set treb_colors(buttonface) #bfbfbf
            set treb_colors(window)     #dfdfdf
            set treb_colors(windowtext) #000000
            set treb_save_file [file join $treb_prefs_dir "Trebuchet Data"]
            set treb_temp_dir $treb_root_dir
            tk_setPalette $treb_colors(window)
            event add <<Cut>>   <Command-Key-x> 
            event add <<Copy>>  <Command-Key-c> 
            event add <<Paste>> <Command-Key-v> 
            event add <<ContextMenu>> <Control-Button-1> <Button-2>
            option add "*Text*relief" "sunken" widgetDefault
            option add "*Entry*relief" "sunken" widgetDefault
            option add "*Text*background" "white" 60
            option add "*Entry*background" "white" 60
            option add "*Listbox*background" "white" 60
            option add "*Listbox*selectBackground" "#cfcfff" 60
            option add "*Text*borderWidth" 2 widgetDefault
            option add "*Text*highlightThickness" 1 widgetDefault
            option add "*Entry*highlightThickness" 1 widgetDefault
            option add "*Checkbutton*highlightThickness" 1 widgetDefault
            option add "*Checkbutton*highlightColor" "#cfcfff" 60
            # Workaround for a Tk8.4a4 bug on mac.
            catch {namespace eval tk set Priv(repeated) 0}
        }
        default {
            set treb_colors(buttonface) [. cget -bg]
            option add *HighlightBackground [. cget -bg]

            entry .flee
            set treb_colors(window)     [.flee cget -bg]
            set treb_colors(windowtext) [.flee cget -fg]
            destroy .flee

            set treb_save_file [file join $treb_prefs_dir ".trebtkrc"]
            set treb_temp_dir "/tmp"
            if {[catch {event add <<ContextMenu>> <Button-3> <Key-App>}]} {
                event add <<ContextMenu>> <Button-3>
            }
            bind Text <Control-Key-comma>   { text_line_shift -4 %W }
            bind Text <Control-Key-period>  { text_line_shift  4 %W }
            bind Text <Control-Key-less>    { text_line_shift -4 %W }
            bind Text <Control-Key-greater> { text_line_shift  4 %W }

            bind Text <Control-Key-v> {}
            event add <<Cut>>   <Control-Key-x> 
            event add <<Copy>>  <Control-Key-c> 
            event add <<Paste>> <Control-Key-v> 

            bind Text    <Button-4> {%W yview scroll -4 units}
            bind Text    <Button-5> {%W yview scroll  4 units}
            bind Listbox <Button-4> {%W yview scroll -4 units}
            bind Listbox <Button-5> {%W yview scroll  4 units}
            bind Canvas  <Button-4> {%W yview scroll -4 units}
            bind Canvas  <Button-5> {%W yview scroll  4 units}
        }
    }

    catch {
        bind Text <MouseWheel> {
            if {%D > 0} {
                %W yview scroll -4 units
            } else {
                %W yview scroll 4 units
            }
        }
        bind Listbox <MouseWheel> {
            if {%D > 0} {
                %W yview scroll -4 units
            } else {
                %W yview scroll 4 units
            }
        }
        bind Canvas <MouseWheel> {
            if {%D > 0} {
                %W yview scroll -4 units
            } else {
                %W yview scroll 4 units
            }
        }
    }

    bind Listbox <Key> {gdm:ListBox:Keypress %W %K %A ; break}
    bind Listbox <Key-Shift_L> {continue}
    bind Listbox <Key-Shift_R> {continue}
    bind Listbox <Key-Control_L> {continue}
    bind Listbox <Key-Control_R> {continue}
    bind Listbox <Key-Alt_L> {continue}
    bind Listbox <Key-Alt_R> {continue}
    bind Listbox <Alt-Key> {continue}
    bind Listbox <Key-Tab> {continue}
    bind Listbox <Shift-Key-Tab> {continue}
    bind Listbox <Control-Key-Tab> {continue}
    bind Listbox <Control-Shift-Key-Tab> {continue}

    bind Text <<Beep>> {/bell ; break}
    event add <<Beep>> <Control-Key-g>

    bind Text <<Cut>> {editCut %W ; break}
    bind Text <<Copy>> {editCopy %W ; break}
    bind Text <<Paste>> {editPaste %W ; break}
    bind Text <<PasteAt>> {editPasteAt %W %x %y ; break}

    bind Entry <<Cut>> {editCut %W ; break}
    bind Entry <<Copy>> {editCopy %W ; break}
    bind Entry <<Paste>> {editPaste %W ; break}
    bind Entry <<PasteAt>> {editPasteAt %W %x %y ; break}

    set treb_lib_dir [file join $treb_root_dir lib]
    set treb_docs_dir [file join $treb_root_dir docs]
    set treb_tool_dir [file join $treb_root_dir pkgs]
    set treb_cacerts_dir [file join $treb_root_dir cacerts]

    image create photo ssl_icon_secure -file [file join $treb_lib_dir images locked.gif]
    image create photo ssl_icon_insecure -file [file join $treb_lib_dir images unlocked.gif]

    source [file join $treb_lib_dir bitmaps.tcl]
    source [file join $treb_lib_dir errors.tcl]
    source [file join $treb_lib_dir spinner.tcl]
    source [file join $treb_lib_dir groupbox.tcl]
    source [file join $treb_lib_dir combobox.tcl]
    source [file join $treb_lib_dir notebook.tcl]
    source [file join $treb_lib_dir tree.tcl]
    source [file join $treb_lib_dir colorwheel.tcl]
    source [file join $treb_lib_dir btnbar.tcl]
    source [file join $treb_lib_dir textpups.tcl]
    source [file join $treb_lib_dir selector.tcl]
    source [file join $treb_lib_dir editdlog.tcl]
    source [file join $treb_lib_dir textdlog.tcl]
    source [file join $treb_lib_dir finddlog.tcl]
    source [file join $treb_lib_dir prefs.tcl]
    source [file join $treb_lib_dir commands.tcl]
    source [file join $treb_lib_dir secsupp.tcl]
    source [file join $treb_lib_dir menus.tcl]
    source [file join $treb_lib_dir mcpmgr.tcl]

    if {![info exists treb_version_c] || $treb_version != $treb_version_c} {
        source [file join $treb_lib_dir compat.tcl]
    }

    calculate_clicks_per_second
    mcp_initialize /socket:sendln_raw
    prefs:init
}


if {[catch {init $argc $argv} mesg]} {
    global errorInfo
    set savedInfo $errorInfo

    set top [toplevel .tle]
    wm title $top "Trebuchet Error"
    text $top.t -font {Courier 10} -yscrollcommand {$top.sb set}
    scrollbar $top.sb -command {$top.t yview} -orient vert
    button $top.b -text "Ok" -width 6 -command {exit} -default active

    grid columnconfigure $top 0 -weight 1
    grid rowconfigure $top 0 -weight 1
    grid rowconfigure $top 1 -minsize 5
    grid $top.t -sticky nsew
    grid $top.sb -row 0 -column 1 -sticky ns
    grid $top.b -row 2 -column 0

    $top.t insert end $savedInfo
    focus $top.b
    tkwait window $top
    exit
}


proc bgerror {mesg} {
    global errorInfo
    set savedInfo $errorInfo
    if {[string match "SSL channel \"*\": *" $mesg]} {
        return ""
    }
    /error [/socket:current] $mesg $savedInfo
    /socket:setforeground
    return ""
}


proc gdm:ListBox:Keypress {wname keycode key} {
    global gdmListBox
    set mylist [$wname get 0 end]
    set lastpos [$wname index anchor]

    if {$keycode == "Up"} {
        set newpos $lastpos
        incr newpos -1
        if {$newpos < 0} {
            set newpos 0
        }
        if {[info exists gdmListBox($wname,typedchars)]} {
            unset gdmListBox($wname,typedchars)
        }
        if {[info exists gdmListBox($wname,keytimer)]} {
            after cancel $gdmListBox($wname,keytimer)
            unset gdmListBox($wname,keytimer)
        }
    } elseif {$keycode == "Down"} {
        set newpos $lastpos
        incr newpos
        if {$newpos > [$wname index end]} {
            set newpos [$wname index end]
        }
        if {[info exists gdmListBox($wname,typedchars)]} {
            unset gdmListBox($wname,typedchars)
        }
        if {[info exists gdmListBox($wname,keytimer)]} {
            after cancel $gdmListBox($wname,keytimer)
            unset gdmListBox($wname,keytimer)
        }
    } else {
        if {![regexp -nocase -- {[ -~]} $key]} {
            if {[info exists gdmListBox($wname,typedchars)]} {
                unset gdmListBox($wname,typedchars)
            }
            if {[info exists gdmListBox($wname,keytimer)]} {
                after cancel $gdmListBox($wname,keytimer)
                unset gdmListBox($wname,keytimer)
            }
            return
        }
        if {[info exists gdmListBox($wname,keytimer)]} {
            after cancel $gdmListBox($wname,keytimer)
        }
        set gdmListBox($wname,keytimer) [
            after 750 "
                if {\[info exists gdmListBox($wname,typedchars)\]} {
                    unset gdmListBox($wname,typedchars)
                }
            "
        ]
        set chars {}
        if {[info exists gdmListBox($wname,typedchars)]} {
            set chars $gdmListBox($wname,typedchars)
        }
        if {$lastpos == {}} {
            set lastpos 0
        }
        append chars $key
        set chars [string tolower $chars]

        set llen [llength $mylist]
        set slen [string length $chars]
        incr slen -1
        for {set newpos 0} {$newpos < $llen} {incr newpos} {
            set item [lindex $mylist $newpos]
            set posstr [string tolower [string range $item 0 $slen]]
            set test  [string compare $posstr $chars]
            if  {$test == 0} {
                break
            }
        }
        if {$newpos >= $llen} {
            set newpos $lastpos
        }
        set gdmListBox($wname,typedchars) $chars
    }

    if {$newpos != $lastpos} {
        $wname selection clear 0 end
        $wname selection set $newpos
        $wname activate $newpos
        $wname see $newpos
    }
}

proc redirect_keystrokes {wout} {
    #bindtags $wout [list [winfo toplevel $wout] $wout Text all]

    global tcl_platform

    # These are the keys which should get redirected to the input buffer.
    # ie: almost everything but Alt, Shift, Control, and Capslock
    set keysyms {
        F1 F2 F3 F4 F5 F6 F7 F8 F9 F10 L1 L2 Escape
        quoteleft 1 2 3 4 5 6 7 8 9 0 minus equal BackSpace Tab
        q w e r t y u i o p bracketleft bracketright Return
        a s d f g h j k l semicolon quoteright
        z x c v b n m comma period slash backslash space
        asciitilde exclam at numbersign dollar percent asciicircum ampersand
        asterisk parenleft parenright underscore plus
        Q W E R T Y U I O P braceleft braceright
        A S D F G H J K L colon quotedbl
        Z X C V B N M less greater question bar
        Insert Delete Home End Prior Next Left Down Right Up
    }
    # These are the control-keys which should get redirected to the input buffer.
    # ie: almost everything but Ctrl-c, Ctrl-x, and modifier keys
    set ctrl_keysyms {
        F1 F2 F3 F4 F5 F6 F7 F8 F9 F10 L1 L2 Escape
        quoteleft 1 2 3 4 5 6 7 8 9 0 minus equal BackSpace Tab
        q w e r t y u i o p bracketleft bracketright Return
        a s d f g h j k l semicolon quoteright
        z     v b n m comma period slash backslash space
        asciitilde exclam at numbersign dollar percent asciicircum ampersand
        asterisk parenleft parenright underscore plus
        Q W E R T Y U I O P braceleft braceright
        A S D F G H J K L colon quotedbl
        Z     V B N M less greater question bar
        Insert Delete Home End Prior Next Left Down Right Up
    }
    if {$tcl_platform(platform) == "macintosh" || $tcl_platform(os) == "Darwin"} {
        # On a macintosh, bind all non-meta keys.
        set ctrl_keysyms $keysyms
    }
    foreach ksym $keysyms {
        bind $wout <Key-$ksym> "
            focus \[/inbuf\]
            event generate \[/inbuf\] <Key-$ksym> -keycode %k
            break
        "
        bind $wout <Shift-Key-$ksym> "
            focus \[/inbuf\]
            event generate \[/inbuf\] <Shift-Key-$ksym> -keycode %k
            break
        "
        bind $wout <Control-Shift-Key-$ksym> "
            focus \[/inbuf\]
            event generate \[/inbuf\] <Control-Shift-Key-$ksym> -keycode %k
            break
        "
        bind $wout <Alt-Key-$ksym> "
            focus \[/inbuf\]
            event generate \[/inbuf\] <Alt-Key-$ksym> -keysym %K -keycode %k
            break
        "
        bind $wout <Alt-Shift-Key-$ksym> "
            focus \[/inbuf\]
            event generate \[/inbuf\] <Alt-Shift-Key-$ksym> -keysym %K -keycode %k
            break
        "
        bind $wout <Alt-Control-Key-$ksym> "
            focus \[/inbuf\]
            event generate \[/inbuf\] <Alt-Control-Key-$ksym> -keysym %K -keycode %k
            break
        "
        bind $wout <Alt-Control-Shift-Key-$ksym> "
            focus \[/inbuf\]
            event generate \[/inbuf\] <Alt-Control-Shift-Key-$ksym> -keysym %K -keycode %k
            break
        "
    }
    foreach ksym $ctrl_keysyms {
        bind $wout <Control-Key-$ksym> "
            focus \[/inbuf\]
            event generate \[/inbuf\] <Control-Key-$ksym> -keycode %k
            break
        "
    }
    bind $wout <<Cut>> "
        event generate $wout <<Copy>>
        break
    "
    bind $wout <<Paste>> {
        event generate [/inbuf] <<Paste>>
        break
    }
    bind $wout <<PasteAt>> {
        event generate [/inbuf] <<Paste>>
        break
    }
    return
}

proc getColorVals {color} {
    return [winfo rgb . $color]
}

proc setvars {variables values} {
    foreach var $variables val $values {
        if {$var == ""} {
            break
        }
        upvar $var tmp
        set tmp $val
    }
    return $values
}

proc dispatcher {root opt argset} {
    if {[info procs $root:$opt] != {}} {
        set cmd "$root:$opt"
        foreach arg $argset {
            append cmd " [list $arg]"
        }
        return [eval "$cmd"]
    } else {
        set opts ""
        set rootlen [string length $root]
        incr rootlen
        foreach item [lsort -dictionary [info procs $root:*]] {
            if {$opts != ""} {
                append opts ", "
            }
            append opts [string range $item $rootlen end]
        }
        error "$root: Unknown option \"$opt\" should be one of $opts"
    }
}

proc lrotate {list first last count} {
    if {$first == "end"} {set first [expr {[llength $list] - 1}]}
    if {$last == "end"} {set last [expr {[llength $list] - 1}]}
    if {$first > $last} {
        set tmp $first
        set first $last
        set last $tmp
        set count [expr {-1 * $count}]
    }
    if {(abs($last - $first) + 1) > $count} {
        set count [expr {$count % (abs($last - $first) + 1)}]
    }
    if {$count > 0} {
        set endfirst [expr {$last - $count}]
        set startlast [expr {$last - ($count - 1)}]
    } elseif {$count < 0} {
        set endfirst [expr {$first - ($count + 1)}]
        set startlast [expr {$first - $count}]
    } else {
        return $list
    }
    set tmprng [lrange $list $startlast $last]
    set tmplist [lreplace $list $startlast $last]
    foreach item $tmprng {
        set tmplist [linsert $tmplist $first $item]
        incr first
    }
    return $tmplist
}

proc lrottostart {list pos} {
    set item [lindex $list $pos]
    return [linsert [lreplace $list $pos $pos] 0 $item]
}

proc chooseColor {args} {
    set results [eval tk_chooseColor $args]
    if {[string match "after#*" $results]} {
        regexp {after#[0-9]+([^0-9].*$)} "$results" {} results
    }
    return $results
}

proc editCopy {w} {
    if {[winfo class $w] == "Text"} {
        if {![catch {set data [$w get sel.first sel.last]}]} {
            if {$data != {}} {
                clipboard clear -displayof $w
                clipboard append -displayof $w -- $data
            }
        }
    } elseif {[winfo class $w] == "Entry"} {
        if {![catch {set data [selection get -displayof $w -selection PRIMARY]}]} {
            if {$data != {}} {
                clipboard clear -displayof $w
                clipboard append -displayof $w -- $data
            }
        }
    }
}

proc editCut {w} {
    if {$w == [/display]} {
        /statbar 5 "You can't cut text from the output display."
        /bell
        return
    }
    if {[winfo class $w] == "Text"} {
        if {![catch {set data [$w get sel.first sel.last]}]} {
            clipboard clear -displayof $w
            clipboard append -displayof $w -- $data
            $w delete sel.first sel.last
        }
    } elseif {[winfo class $w] == "Entry"} {
        if {![catch {set data [selection get -displayof $w -selection PRIMARY]}]} {
            clipboard clear -displayof $w
            clipboard append -displayof $w -- $data
            $w delete sel.first sel.last
        }
    }
}

proc editPaste {w} {
    if {$w == [/display]} {
        set w [/inbuf]
        #/statbar 5 "You can't paste into the output display."
        #/bell
        #return
    }
    catch {
        catch {
            $w delete sel.first sel.last
        }
        set data [selection get -displayof $w -selection CLIPBOARD]
        $w insert insert $data
    }
}

proc editPasteAt {w x y} {
    if {$w == [/display]} {
        /statbar 5 "You can't paste into the output display."
        /bell
        return
    }
    catch {
        if {[winfo class $w] == "Text"} {
            catch { $w tag remove sel.first sel.last }
        } elseif {[winfo class $w] == "Entry"} {
            catch { $w selection clear }
        }
        $w insert @$x,$y [selection get -displayof $w -selection CLIPBOARD]
        if {[winfo class $w] == "Text"} {
            $w mark set insert @%x,%y
        } elseif {[winfo class $w] == "Entry"} {
            $w icursor @%x,%y
        }
    }
}

proc match_text_word {wname pos {end {}}} {
    global nonwordchar
    set first [$wname search -backwards -regexp -nocase -- "$nonwordchar|^" "$pos + 1 chars"]
    if {[$wname compare $first != "$first linestart"]} {
        set first "$first + 1 chars"
    }

    if {$end != {}} {
        set last $end
    } else {
        set last [$wname search -forwards -regexp -nocase -- "$nonwordchar|$" $pos]
    }
    return [$wname get $first "$last"]
}

proc complete_word {wname word} {
    global nonwordchar wordchar
    regsub -all {[]\$.*()|?+^; []} $word {\\&} word
    set limindex "end - [/prefs:get completion_lines] lines"
    set lastindex "end"
    set shortword ""
    set shortlen 0
    set wordslist {}
    set wordlen [string length $word]
    while {$lastindex != ""} {
        set lastindex [$wname search -backwards -regexp -nocase -- \
            "$nonwordchar$word|^$word" $lastindex $limindex]
        if {$lastindex != ""} {
            if {[$wname compare $lastindex != "$lastindex linestart"]} {
                set wordstart "$lastindex + 1 chars"
            } else {
                set wordstart "$lastindex"
            }
            set foundword [match_text_word $wname "$wordstart"]
            if {$shortword == ""} {
                set shortword $foundword
                set shortlen [string length $shortword]
                lappend wordslist [string tolower "$foundword"]
            } else {
                if {[lsearch -exact "$wordslist" [string tolower "$foundword"]] == -1} {
                    lappend wordslist [string tolower "$foundword"]
                }
                set tmpword [string range $foundword 0 [expr {$shortlen - 1}]]
                while {[string tolower $shortword] != [string tolower $tmpword]} {
                    incr shortlen -1
                    set tmplen [expr {$shortlen - 1}]
                    set tmpword [string range $tmpword 0 $tmplen]
                    set shortword [string range $shortword 0 $tmplen]
                }
            }
            set lastindex "$lastindex - 1 chars"
        }
    }
    return [list $shortword [lsort -dictionary $wordslist]]
}

proc complete_word_proc {} {
    global wordchar nonwordchar
    /statbar 0 {}
    set matchedword [match_text_word [/inbuf] "insert - 1 chars" "insert"]
    if {![regexp -nocase -- "$wordchar+" "$matchedword"]} {
        /bell
        /statbar 5 "No word to complete."
        return ""
    }
    set results [complete_word [/display] $matchedword]
    set newword [lindex $results 0]
    if {$newword != ""} {
        [/inbuf] delete "insert - [string length $matchedword] chars" insert
        [/inbuf] insert insert "$newword"
        set wordslist [lindex $results 1]
        if {[llength $wordslist] > 1} {
            /bell
            /statbar 5 "Ambiguous: $wordslist"
        } else {
            [/inbuf] insert insert " "
        }
    } else {
        /bell
        /statbar 5 "No matches for '$matchedword'"
    }
    return ""
}


proc init_menus {window} {
    init_main_menus $window
    return ""
}

proc /show_copyright {} {
    global treb_root_dir
    /textdlog -buttons -title "GNU General Public License" \
        -width 80 -height 24 -nowrap -readonly \
        -file [file join $treb_root_dir COPYING]
}

proc /show_about {} {
    global treb_version treb_name
    tk_messageBox -default ok -type ok -title {About Trebuchet} -message \
        "$treb_name  $treb_version\nCopyright 1998-2003 by Fuzzball Software\nReleased under the GNU Public License\n\n\"An Excellent Way to Sling Mud!\""
    focus [/inbuf]
    return ""
}

proc match_by_name {name list} {
    set cnt 0
    foreach item "$list" {
        if {[lindex $item 0] == $name} {
            return $cnt
        }
        incr cnt
    }
    return -1
}


proc process_command {text {socket {}}} {
    global widget errorInfo errorCode

    if {$socket == {}} {
        set socket [/socket:current]
        set oldsocket {}
    } else {
        set oldsocket [/socket:current]
        if {$oldsocket == $socket} {
            set oldsocket {}
        } else {
            /socket:setcurrent $socket
        }
    }
    set statbar $widget(statbar)
    set inbuf [/inbuf]
    set disp [/display]
    set result ""

    if {[string range "$text" 0 0] == "/"} {
        if {[string length $text] == 1 || [string range "$text" 0 1] == "/ "} {
            # Ignore line as a comment.
        } elseif {[string range "$text" 0 1] != "//"} {
            set cmd [string range "$text" 1 end]
            set pos [string first " " $cmd]
            if {$pos > 0} {
                set macroname [string range $cmd 0 [expr {$pos-1}]]
                set restofline [string range $cmd [expr {$pos+1}] end]
            } else {
                set macroname $cmd
                set restofline ""
            }
            if {[/macro:exists $macroname]} {
                set ret [catch {/macro:execute $macroname $restofline} result]
                if {$ret != 0} {
                    set savedInfo $errorInfo
                    if {$oldsocket != {}} {
                        /socket:setcurrent $oldsocket
                    }
                    set lines [split $savedInfo "\n"]
                    set cnt [llength $lines]
                    incr cnt -4
                    set savedInfo [join [lrange $lines 0 $cnt] "\n"]
                    error $result $savedInfo $errorCode
                }
            } else {
                set ret [catch {uplevel #0 "/$cmd"} result]
                if {$ret != 0} {
                    set savedInfo $errorInfo
                    if {$oldsocket != {}} {
                        /socket:setcurrent $oldsocket
                    }
                    set lines [split $savedInfo "\n"]
                    set cnt [llength $lines]
                    incr cnt -4
                    set savedInfo [join [lrange $lines 0 $cnt] "\n"]
                    error $result $savedInfo $errorCode
                }
            }
        } else {
            /socket:sendln $socket [string range $text 1 end]
        }
    } elseif {[string range "$text" 0 0] == "\\"} {
        if {[string range "$text" 0 1] != "\\\\"} {
            set cmd [string range "$text" 1 end]
            set ret [catch {uplevel #0 "/socket:sendln [list $socket] \"$cmd\""} result]
            if {$ret != 0} {
                set savedInfo $errorInfo
                if {$oldsocket != {}} {
                    /socket:setcurrent $oldsocket
                }
                set lines [split $savedInfo "\n"]
                set cnt [llength $lines]
                incr cnt -4
                set savedInfo [join [lrange $lines 0 $cnt] "\n"]
                error $result $savedInfo $errorCode
            }
            set result ""
        } else {
            /socket:sendln $socket [string range $text 1 end]
        }
    } else {
        /socket:sendln $socket "$text"
    }
    if {$oldsocket != {}} {
        /socket:setcurrent $oldsocket
    }
    return $result
}

proc process_commands {text {socket {}}} {
    global errors_nonfatal errorInfo errorCode
    if {$socket == {}} {
        set socket [/socket:current]
        set oldsocket {}
    } else {
        set oldsocket [/socket:current]
        if {$oldsocket == $socket} {
            set oldsocket {}
        } else {
            /socket:setcurrent $socket
        }
    }
    if {[string first "\n" $text] >= 0} {
        set cmds [split $text "\n"]
        set ccmd ""
        foreach cmd $cmds {
            if {$ccmd != ""} {
                append ccmd "\n"
            }
            append ccmd $cmd
            if {[info complete $ccmd]} {
                set ret [catch {process_command $ccmd $socket} result]
                if {$ret != 0} {
                    set savedInfo $errorInfo
                    if {!$errors_nonfatal} {
                        if {$oldsocket != {}} {
                            /socket:setcurrent $oldsocket
                        }
                        # set lines [split $savedInfo "\n"]
                        # set cnt [llength $lines]
                        # incr cnt -4
                        # set savedInfo [join [lrange $lines 0 $cnt] "\n"]
                        return -code $ret -errorinfo $savedInfo -errorcode $errorCode $result
                        error $result $savedInfo $errorCode
                    }
                    /nonmodalerror $socket $errMsg
                }
                set ccmd ""
            }
        }
        if {$ccmd != ""} {
            set ret [catch {process_command $ccmd $socket} result]
            if {$ret != 0} {
                set savedInfo $errorInfo
                if {!$errors_nonfatal} {
                    if {$oldsocket != {}} {
                        /socket:setcurrent $oldsocket
                    }
                    # set lines [split $savedInfo "\n"]
                    # set cnt [llength $lines]
                    # incr cnt -4
                    # set savedInfo [join [lrange $lines 0 $cnt] "\n"]
                    return -code $ret -errorinfo $savedInfo -errorcode $errorCode $result
                    error $result $savedInfo $errorCode
                }
                /nonmodalerror $socket $errMsg
            }
            set ccmd ""
        }
    } else {
        set ret [catch {process_command $text $socket} result]
        if {$ret != 0} {
            set savedInfo $errorInfo
            if {!$errors_nonfatal} {
                if {$oldsocket != {}} {
                    /socket:setcurrent $oldsocket
                }
                # set lines [split $savedInfo "\n"]
                # set cnt [llength $lines]
                # incr cnt -4
                # set savedInfo [join [lrange $lines 0 $cnt] "\n"]
                return -code $ret -errorinfo $savedInfo -errorcode $errorCode $result
                error $result $savedInfo $errorCode
            }
            /nonmodalerror $socket $errMsg
        }
    }
    if {$oldsocket != {}} {
        /socket:setcurrent $oldsocket
    }
    return $result
}

proc proxyconnect {sok world} {
    upvar #0 proxy_delay_connect delay_connect
    set delay_connect 0

    fileevent $sok writable ""
    /socket:setcurrent $world

    global tcl_platform
    if {![catch {fconfigure $sok -error} sokerr]} {
        if {$sokerr != {} } {
            /socket:disconnect $world
            /statbar 10 "Connection to $world failed."
            update idletasks
            /error $world "Could not connect to world $world.\n$sokerr"
            return
        }
    }

    set host [/world:get host $world]
    set port [/world:get port $world]

    if {[/macro:exists proxy_command]} {
        if {[catch {/macro:execute -w $world proxy_command [list $host] [list $port]} errMsg] && $errMsg != ""} {
            /nonmodalerror "" $errMsg
        }
        if {[/macro:exists proxy_read]} {
            fileevent $sok readable "proxyread $sok [list $world]"
        } elseif {[info exists delay_connect] && $delay_connect} {
            fileevent $sok readable "delayconnect $sok [list $world]"
        } else {
            fileevent $sok readable "readline [list $world]"
            directconnect $sok $world
        }
    } else {
        fileevent $sok readable "readline [list $world]"
        directconnect $sok $world
    }
    unset delay_connect
}

proc proxyread {sok world} {
    global tcl_platform
    if {![catch {fconfigure $sok -error} sokerr]} {
        if {$sokerr != {} } {
            /socket:disconnect $world
            /statbar 10 "Connection to $world failed."
            update idletasks
            /error $world "Could not connect to world $world.\n$sokerr"
            return
        }
    }

    upvar #0 proxy_connected connected
    upvar #0 proxy_delay_connect delay_connect
    upvar #0 proxy_error prerror
    set connected 0
    set delay_connect 0
    set prerror ""
    if {[/macro:exists proxy_read]} {
        /socket:setcurrent $world
        if {[catch {/macro:execute proxy_read "[gets $sok]"} errMsg] && $errMsg != ""} {
            fileevent $sok readable "readline [list $world]"
            directconnect $sok $world
            /nonmodalerror $world $errMsg
            return
        }
        if {[info exists prerror] && $prerror != ""} {
            /statbar 10 "Connection to $world failed."
            /socket:disconnect $world
            unset connected
            unset delay_connect
            unset prerror
            error "Connect: $prerror"
        }
        if {[info exists delay_connect] && $delay_connect} {
            fileevent $sok readable "delayconnect $sok [list $world]"
        } elseif {[info exists connected] && $connected} {
            fileevent $sok readable "readline [list $world]"
            directconnect $sok $world
        }
    } else {
        fileevent $sok readable "readline [list $world]"
        directconnect $sok $world
        readline $world
    }
    unset connected
    unset delay_connect
}

proc delayconnect {sok world} {
    directconnect $sok $world
    readline $world
}


proc directconnect {sok world} {
    fileevent $sok writable ""

    global tcl_platform
    if {![catch {fconfigure $sok -error} sokerr]} {
        if {$sokerr != {} } {
            /socket:disconnect $world
            /statbar 10 "Connection to $world failed."
            update idletasks
            /error $world "Could not connect to world $world.\n$sokerr"
            return
        }
    }

    if {[info commands tls::init] != {}} {
        if {[/world:get secure $world]} {
            /statbar 30 "Negotiating SSL for $world..."
            update idletasks
            global treb_cacerts_dir
            #tls::import $sok -require 1 -cadir $treb_cacerts_dir -ssl2 0 -tls1 1 -command tls_command_cb
            tls::import $sok -require 1 -cadir $treb_cacerts_dir -ssl2 1 -tls1 1 -command tls_command_cb
            wait_for_ssl_negotiation $sok $world
            return
        }
    }
    connect_complete $sok $world
}

proc connect_complete {sok world} {
    fileevent $sok writable ""
    if {[info commands tls::init] != {}} {
        if {[/world:get secure $world]} {
            if {[catch {tls::status $sok} result]} {
                after 250 "catch {fileevent [list $sok] writable \"connect_complete [list $sok] [list $world]\"}"
                return
            } elseif {$result == {}} {
                after 250 "catch {fileevent [list $sok] writable \"connect_complete [list $sok] [list $world]\"}"
                return
            }
        }
    }

    /socket:set state $world "Connected"
    /statbar 5 "Connection to $world established..."

    set char [/world:get charname $world]
    set pass [/world:get password $world]

    if {$char != {}} {
        if {[/world:get type $world] == "tiny"} {
            /socket:sendln $world "connect $char $pass"
        } elseif {[/world:get type $cworld] == "lp"} {
            /socket:sendln $world "$char"
            /socket:sendln $world "$pass"
        } elseif {[/world:get type $cworld] == "lpp"} {
            /socket:sendln $world "$char"
            /socket:sendln $world "$pass"
        } else {
            /socket:sendln $world "connect $char $pass"
        }
        if {[/world:get tcl $world]} {
            eval [/world:get script $world]
        } else {
            process_commands [/world:get script $world] $world
        }
    }
    /socket:setlight $world green
}

proc readline {world {forced_data ""}} {
    global treb_readlines
    global treb_partial_line
    global treb_partial_line_timer

    hilite:update_match_proc

    set sok [/socket:get socket $world]

    if {![info exists treb_partial_line($sok)]} {
        set treb_partial_line($sok) {}
    }

    if {$forced_data != ""} {
        set newdata $forced_data
    } else {
        if {[info commands tls::init] != {}} {
            if {[/world:get secure $world]} {
                if {[catch {tls::status $sok} result]} {
                    fileevent $sok readable ""
                    after 250 "catch {fileevent [list $sok] readable \"readline [list $world]\"}"
                    return
                } elseif {$result == {}} {
                    fileevent $sok readable ""
                    after 250 "catch {fileevent [list $sok] readable \"readline [list $world]\"}"
                    return
                }
            }
        }
        if {![catch {fconfigure $sok -error} sokerr]} {
            if {$sokerr != {} } {
                if {[/socket:get state $world] == "Connected"} {
                    /socket:disconnect $world
                    /statbar 10 "Disconnected from $world."
                    update idletasks
                    /error $world "Your connection to world $world has been lost.\n$sokerr"
                } else {
                    /socket:disconnect $world
                    /statbar 10 "Connection to $world failed."
                    update idletasks
                    /error $world "Could not connect to world $world.\n$sokerr"
                }
                return
            }
        }
        if {[eof $sok]} {
            /socket:disconnect $world
            return
        }
        set newdata "[read $sok]"
    }

    global treb_telnet_enabled

    # Telnet Protocol commands.
    set IAC  "\377" ;# Telnet escape character
    set DONT "\376"
    set DO   "\375"
    set WONT "\374"
    set WILL "\373"
    set GA   "\371" ;# Go Ahead
    set AYT  "\366" ;# Are You There
    set EOR  "\357" ;# End Of Record

    # Telnet Protocol capabilities.
    set END_OF_RECORD  "\031" ;# End Of Record capability

    while {[set pos [string first $IAC $newdata]] != -1} {
        append treb_partial_line($sok) [string range $newdata 0 [expr {$pos-1}]]
        set cmd [string index $newdata [expr {$pos+1}]]
        if {$cmd == "$DO"} {
            set arg [string index $newdata [expr {$pos+2}]]
            set newdata [string range $newdata [expr {$pos+3}] end]
            puts -nonewline $sok "$IAC$WONT$arg"
            flush $sok
            set treb_telnet_enabled($sok) 1
        } elseif {$cmd == "$DONT"} {
            set arg [string index $newdata [expr {$pos+2}]]
            set newdata [string range $newdata [expr {$pos+3}] end]
            puts -nonewline $sok "$IAC$WONT$arg"
            flush $sok
            set treb_telnet_enabled($sok) 1
        } elseif {$cmd == "$WILL"} {
            set arg [string index $newdata [expr {$pos+2}]]
            set newdata [string range $newdata [expr {$pos+3}] end]
            if {$arg == "$END_OF_RECORD"} {
                puts -nonewline $sok "$IAC$DO$arg"
            } else {
                puts -nonewline $sok "$IAC$DONT$arg"
            }
            flush $sok
            set treb_telnet_enabled($sok) 1
        } elseif {$cmd == "$WONT"} {
            set newdata [string range $newdata [expr {$pos+3}] end]
            set treb_telnet_enabled($sok) 1
            # ignore this.
        } elseif {$cmd == "$AYT"} {
            set newdata [string range $newdata [expr {$pos+2}] end]
            puts -nonewline $sok {[Yes]}
            flush $sok
        } elseif {$cmd == $EOR} {
            append treb_partial_line($sok) "\n"
            set newdata [string range $newdata [expr {$pos+2}] end]
        } elseif {$cmd == $GA} {
            append treb_partial_line($sok) "\n"
            set newdata [string range $newdata [expr {$pos+2}] end]
        } elseif {$cmd == $IAC} {
            append treb_partial_line($sok) $IAC
            set newdata [string range $newdata [expr {$pos+2}] end]
        } else {
            set newdata [string range $newdata [expr {$pos+2}] end]
        }
    }
    append treb_partial_line($sok) $newdata
    if {![info exists treb_partial_line_timer($sok)]} {
        set treb_partial_line_timer($sok) 0
    }

    regsub -all -- "\r" $treb_partial_line($sok) "" treb_partial_line($sok)
    set lines [split $treb_partial_line($sok) "\n"]
    set linecount [expr {[llength $lines] - 1}]
    set treb_partial_line($sok) [lindex $lines $linecount]

    for {set iline 0} {$iline< $linecount} {incr iline} {
        display:clearpartial [/display $world]
        set treb_partial_line_timer($sok) 0
        set line [lindex $lines $iline]

        if {$line != ""} {
            if {[catch {set line [mcp_process_input $world $line]} errMesg]} {
                /statbar 5 $errMesg
            }
            if {$line == {}} continue
        }
        if {[info exists treb_readlines(pattern,$world)]} {
            set pat $treb_readlines(pattern,$world)
            set cmd $treb_readlines(command,$world)
            if {$pat != ""} {
                if {[string match $pat [string tolower $line]]} {
                    if {!$treb_readlines(nolast,$world)} {
                        lappend treb_readlines(text,$world) $line
                    }
                    set cmd [/line_subst $cmd $treb_readlines(text,$world)]
                    unset treb_readlines(pattern,$world)
                    unset treb_readlines(command,$world)
                    unset treb_readlines(text,$world)
                    unset treb_readlines(nolast,$world)
                    /socket:setcurrent $world
                    if {[catch {eval $cmd} errMsg]} {
                        global errorInfo
                        /error $world $errMsg $errorInfo
                    }
                } else {
                    lappend treb_readlines(text,$world) $line
                }
                continue
            }
        }
        if {[string first "\a" "$line"] != -1} {
            /bell
            regsub -all -- "\a" $line "" line
        }
        /hilite:processline $world $line
    }
    return
}

proc update_charcount {} {
    set pos [split [[/inbuf] index insert] .]
    set line L[lindex $pos 0]
    set char C[lindex $pos 1]
    set lastline [.mw.statusbar.pos.line cget -text]
    set lastchar [.mw.statusbar.pos.char cget -text]
    if {$line != $lastline} {
        set width [.mw.statusbar.pos.line cget -width]
        set len [string length $line]
        if {$len > $width} {
            set width $len
        }
        .mw.statusbar.pos.line configure -text $line -width $width
    }
    if {$char != $lastchar} {
        set width [.mw.statusbar.pos.char cget -width]
        set len [string length $char]
        if {$len > $width} {
            set width $len
        }
        .mw.statusbar.pos.char configure -text $char -width $width
    }
}


set standard_timer_counter 0
proc standard_timer {} {
    global standard_timer_counter
    set count [incr standard_timer_counter]

    global treb_partial_line
    global treb_partial_line_timer
    foreach cworld [/socket:connectednames] {
        catch {
            if {[/world:get type $cworld] == "lp"} {
                set sok [/socket:get socket $cworld]
                if {[info exists treb_partial_line_timer($sok)]} {
                    set timerval [incr treb_partial_line_timer($sok)]
                    if {$timerval == [/prefs:get lpprompt_delay]} {
                        set partline $treb_partial_line($sok)
                        if {$partline != ""} {
                            readline $cworld "\n"
                            # /hilite:processline $cworld $partline 1
                        }
                    }
                }
            }
        }
    }

    hilite:update_match_proc
    update_charcount
    update_secure_indicator
    socket:send_keepalives

    if {$count % 2 == 0} {
        socket:blink_lights
    }
    if {$count % 10 == 0} {
        # Only do this once every 4 timer cycles.
        # (About once a second.)
        /menu:update
    }

    after 100 standard_timer
    return ""
}


proc styleconf {tagname} {
    set oot {}
    foreach i [[/display ""] tag configure $tagname] {
        if {[lindex $i 4] != {}} {
            set oot [concat $oot [lindex $i 0] [list [lindex $i 4]]]
        }
    }
    return $oot
}

global display_number; set display_number 0

proc display:names {} {
    global widget
    return [winfo children $widget(disp)]
}

proc display:current {} {
    global widget
    return [pack slaves $widget(disp)]
}

proc display:at_end {{disp ""}} {
    if {$disp == ""} {
        set disp "[display:current].disp"
    }

    set info [$disp dlineinfo end-1c]
    if {$info == {}} {
        return 0
    }
    # dlineinfo returns clipped size.  there's no easy way to check if the
    # entire line is visible, so instead we see if the baseline is visible.
    if {[lindex $info 3] <= [lindex $info 4]} {
        return 0
    }
    return 1
}

proc display:prev {} {
    global widget
    if {[llength [worldbutton:names]] == 0} {
        display:select $widget(backdrop)
    } else {
        worldbutton:prev
    }
}

proc display:next {} {
    global widget
    if {[llength [worldbutton:names]] == 0} {
        display:select $widget(backdrop)
    } else {
        worldbutton:next
    }
}


proc display:clearpartial {disp} {
    if {$disp == ""} {
        set disp [display:current]
    }
    catch {
        $disp delete "__partialline__.first-1c" "__partialline__.last"
        $disp tag delete "__partialline__"
    }
}


proc display:write {disp text {tagspans {}}} {
    if {$disp == ""} {
        set disp [display:current]
    }

    set at_end [display:at_end $disp]


    $disp mark set "treb_buffer_end" "end-2c"
    $disp insert end "$text" "normal"
    set hfirst [$disp index "treb_buffer_end+1c"]
    foreach attr $tagspans {
        if {$attr != {}} {
            set style [lindex $attr 2]
            set start "$hfirst + [lindex $attr 0] chars"
            set end [lindex $attr 1]
            if {$end == "end"} {
                set end [string length $text]
            }
            set end "$start + $end chars"
            if {[$disp compare $end == "end"]} {
                set end "end-1c"
            }
            $disp tag add $style $start $end
        }
    }

    set lines [expr {int([$disp index end-1c])}]
    set max [/prefs:get scrollback_lines]
    if {$max < $lines} {
        set diff [expr {$lines - $max + 1}]
        $disp delete 1.0 "$diff.0"
    }
    display:autoscroll $disp
    return $at_end
}

proc display:autoscroll {disp} {
    global displayinfo

    if {! [/prefs:get jumpscroll_flag]} {
        display:doautoscroll $disp
    } elseif {$displayinfo($disp,autoscroll_task) == {}} {
        set displayinfo($disp,autoscroll_task) \
            [after 200 display:doautoscroll $disp]
    }
}

proc display:doautoscroll {{disp ""}} {
    global displayinfo

    if {$disp == ""} {
        set disp "[display:current].disp"
    }
    if {![winfo exists $disp]} {
        return 0
    }

    if {$displayinfo($disp,autoscrolling) && ! [display:at_end $disp]} {
        if {[/prefs:get pager_flag]} {
            $disp yview pagestop_mark
        } else {
            $disp see end
        }
    }

    set displayinfo($disp,autoscroll_task) {}
    set displayinfo($disp,autoscrolling) [display:at_end $disp]
}

proc display:markseen {{disp ""}} {
    global displayinfo

    if {$disp == ""} {
        set disp "[display:current].disp"
    }

    set y [expr {[winfo height $disp] - 1}]
    set pos [$disp index "@0,$y linestart"]
    if {[$disp compare pagestop_mark < $pos]} {
        $disp mark set pagestop_mark $pos
    }

    set displayinfo($disp,autoscrolling) [display:at_end $disp]
}

proc display:select {display} {
    global displayinfo widget
    set disp [display:current]
    if {$disp != {}} {
        pack forget $disp
    }
    pack $display -anchor center -expand 1 -fill both -side top
    if {$display != $widget(backdrop)} {
        worldbutton:setpressed $displayinfo($display,worldbutton)
    }
    /socket:update_indicators
}

proc display:setlight {display color} {
    global displayinfo
    set btn $displayinfo($display,worldbutton)
    worldbutton:setlight $btn $color
    return
}

proc display:delete {display} {
    global displayinfo
    worldbutton:delete $displayinfo($display,worldbutton)
    display:next
    destroy $display
}

proc display:mkdisplay {base} {
    global displayinfo treb_fonts tcl_platform
    if {$tcl_platform(os) == "Darwin"} {
        frame $base -relief flat -borderwidth 0
    } else {
        frame $base -relief sunken -borderwidth 2
    }
    set sbarbw 0
    if {$tcl_platform(platform) == "unix"} {
        if {$tcl_platform(os) != "Darwin"} {
            set sbarbw 1
        }
    }
    scrollbar $base.scroll \
        -command "display:markseen $base.disp ; $base.disp yview" -orient vert -borderwidth $sbarbw -relief raised
    text $base.disp \
        -font $treb_fonts(fixed) -wrap word -insertwidth 0 \
        -highlightthickness 0 -relief flat -borderwidth 0 \
        -yscrollcommand "$base.scroll set"
    bind $base.disp <<Copy>> {editCopy %W ; break}
    bind $base.disp <<Paste>> {
        focus [/inbuf]
        event generate [/inbuf] <<Paste>>
        break
    }
    bind $base.disp <<PasteAt>> {
        bell
        break
    }
    if {$tcl_platform(platform) == "unix"} {
        if {$tcl_platform(os) != "Darwin"} {
            bind $base.disp <Button-4> {
                /dokey scroll_line_up
                /dokey scroll_line_up
                /dokey scroll_line_up
                /dokey scroll_line_up
                break;
            }
            bind $base.disp <Button-5> {
                /dokey scroll_line_dn
                /dokey scroll_line_dn
                /dokey scroll_line_dn
                /dokey scroll_line_dn
                break;
            }
        }
    }
    redirect_keystrokes $base.disp
    $base.disp tag add normal 0.0 end

    style:ansi_update $base.disp

    grid columnconfigure $base 1 -weight 0
    grid columnconfigure $base 0 -weight 1
    grid rowconfigure $base 0 -weight 1
    grid $base.scroll -row 0 -column 1 -sticky nsew
    grid $base.disp -row 0 -column 0 -sticky nsew
    textPopup:new $base.disp
    $base.disp mark set pagestop_mark 1.0
    $base.disp mark gravity pagestop_mark left
    set displayinfo($base.disp,autoscrolling) 1
    set displayinfo($base.disp,autoscroll_task) {}
    return
}

proc display:add {label} {
    global display_number displayinfo widget
    incr display_number

    set base $widget(disp).$display_number
    display:mkdisplay $base
    set btn [worldbutton:add $label "display:select $base"]
    set displayinfo($base,worldbutton) $btn
    worldbutton:press $btn
    return $base
}




proc worldbutton:names {} {
    global widget
    return [winfo children $widget(worldsbar)]
}

proc worldbutton:current {} {
    foreach btn [worldbutton:names] {
        if {[$btn cget -relief] == "solid"} {
            return $btn
        }
    }
    return ""
}

proc worldbutton:prev {} {
    set worldbtns [worldbutton:names]
    set pos [lsearch -exact $worldbtns [worldbutton:current]]
    set numbtns [llength $worldbtns]
    set pos [expr {($pos + $numbtns - 1) % $numbtns}]
    worldbutton:press [lindex $worldbtns $pos]
    return
}

proc worldbutton:next {} {
    set worldbtns [worldbutton:names]
    set pos [lsearch -exact $worldbtns [worldbutton:current]]
    set pos [expr {($pos + 1) % [llength $worldbtns]}]
    worldbutton:press [lindex $worldbtns $pos]
    return
}

proc worldbutton:setpressed {which} {
    global widget
    foreach btn [winfo children $widget(worldsbar)] {
        $btn configure -relief groove -borderwidth 2
    }
    $which configure -relief solid -borderwidth 2
    return
}

proc worldbutton:press {which} {
    global worldbutton_info
    worldbutton:setpressed $which
    eval "$worldbutton_info($which,script)"
    return
}

proc worldbutton:setlight {which color} {
    set lightimg [gdm:Bitmap get ${color}light]
    if {$lightimg != [$which.img cget -image]} {
        $which.img config -image $lightimg
    }
    return
}

global world_button_number; set world_button_number 0

proc worldbutton:add {{world ""} {script ""}} {
    global widget world_button_number worldbutton_info
    set base $widget(worldsbar)
    incr world_button_number
    set name $base.world$world_button_number
        set inittitle $world
    if {$world == ""} {
        set world "World $world_button_number"
    }
    set lightimg [gdm:Bitmap get yellowlight]
    frame $name -relief groove -borderwidth 2
    label $name.img -relief flat -image $lightimg -borderwidth 0
    label $name.lbl -relief flat -text "$world" -borderwidth 0
    pack $name.img -anchor w -side left
    pack $name.lbl -anchor w -side left
    pack $name -anchor center -side left -padx 2 -fill none -ipadx 2
    bind $name <Button-1> "worldbutton:press $name"
    bind $name.img <Button-1> "worldbutton:press $name"
    bind $name.lbl <Button-1> "worldbutton:press $name"
    if {$inittitle != ""} {
        set mymenu $name.menu
        menu $mymenu -tearoff 0 
        $mymenu add command -label "Reconnect World" \
            -command "/reconnect [list $world]"
        $mymenu add command -label "Disconnect World" \
            -command "/dc [list $world]"
        $mymenu add command -label "Close World" \
            -command "/close [list $world]"
        bind $name <<ContextMenu>> "tk_popup $mymenu %X %Y"
        bind $name.img <<ContextMenu>> "tk_popup $mymenu %X %Y"
        bind $name.lbl <<ContextMenu>> "tk_popup $mymenu %X %Y"
    }
    set worldbutton_info($name,script) $script
    return $name
}

proc worldbutton:delete {btn} {
    destroy $btn
}


proc gdm:TextUpDownSingleLine {w n} {
    if {[info exists tk::Priv]} {
        upvar #0 tk::Priv priv
    } else {
        upvar #0 tkPriv priv
    }

    $w see insert
    set i [$w index insert]
    set ibox [$w bbox insert]
    set curx [lindex $ibox 0]
    set cury [lindex $ibox 1]
    set oldy $cury
    set curh [lindex $ibox 3]
    set yslop [expr {[$w cget -borderwidth] + [$w cget -highlightthickness] + [$w cget -pady]}]
    if {$n > 0} {
        set xslop [expr {[$w cget -borderwidth] + [$w cget -padx]}]
        set winw [expr {[winfo width $w] - $xslop}]
        set winh [expr {[winfo height $w] - $yslop}]
        if {$cury + $curh >= $winh && [$w compare @$winw,$winh != end-1c]} {
            $w yview scroll 1 units
        } else {
            incr cury $curh
        }
    } elseif {$n < 0} {
        if {$cury - 1 <= $yslop && [$w compare @0,0 != 1.0]} {
            $w yview scroll -1 units
        } else {
            incr cury -1
        }
    }
    scan $i "%d.%d" line char
    if {[string compare $priv(prevPos) $i] != 0} {
        set priv(char) $char
    }
    set new [$w index @$curx,$cury]
    set priv(prevPos) $new
    return $new
}

proc tkTextUpDownLine {w n} {
    set step [expr {abs($n)}]
    if {$n > 0} {
        for {set i $n} {$i > 0} {incr i -1} {
            set new [gdm:TextUpDownSingleLine $w 1]
        }
    } else {
        for {set i $n} {$i < 0} {incr i 1} {
            set new [gdm:TextUpDownSingleLine $w -1]
        }
    }
    return $new
}

if {$tcl_platform(os) == "Darwin"} {
    namespace eval ::tk {
        proc TextUpDownLine {w n} {
            tkTextUpDownLine $w $n
        }
    }
}


#
# Calculates a new color by averaging to two existing colors.
#
proc colorAverage {color1 color2} {
set rgb1 [winfo rgb . $color1]
set red1 [lindex $rgb1 0]
set grn1 [lindex $rgb1 1]
set blu1 [lindex $rgb1 2]

set rgb2 [winfo rgb . $color2]
set red2 [lindex $rgb2 0]
set grn2 [lindex $rgb2 1]
set blu2 [lindex $rgb2 2]

set red [expr {($red1 + $red2) / 2}]
set grn [expr {($grn1 + $grn2) / 2}]
set blu [expr {($blu1 + $blu2) / 2}]

return [format #%04x%04x%04x $red $grn $blu]
}


#
# Calculates a new color, darkened or lightened from an existing color.
#
proc colorShade {color percent} {
set rgb [winfo rgb . $color]
set red [lindex $rgb 0]
set grn [lindex $rgb 1]
set blu [lindex $rgb 2]

set max $red
if {$grn > $max} {set max $grn}
if {$blu > $max} {set max $blu}

set delta [expr {int($max * ($percent - 100) / 100)}]
set red [expr {$red + $delta}]
set grn [expr {$grn + $delta}]
set blu [expr {$blu + $delta}]

if {$red < 0} {set red 0}
if {$grn < 0} {set grn 0}
if {$blu < 0} {set blu 0}

if {$red > 0xffff} {set red 0xffff}
if {$grn > 0xffff} {set grn 0xffff}
if {$blu > 0xffff} {set blu 0xffff}

return [format #%04x%04x%04x $red $grn $blu]
}


proc text_line_shift {chars w} {
    if {[catch {
        set first [$w index "sel.first linestart"]
        set last [$w index "sel.last lineend"]
    }]} {
        set first [$w index "insert linestart"]
        set last [$w index "insert lineend"]
    }
    set curr $first
    while {[$w compare $curr < $last]} {
        set line [$w get $curr "$curr lineend"]
        regexp {^[ 	]*} $line whitespace
        set count [string length $whitespace]
        regsub -all -- {	} $whitespace {        } whitespace
        if {$chars < 0} {
            set whitespace [string range $whitespace [expr {abs($chars)}] end]
        } else {
            for {set i 0} {$i < $chars} {incr i} {
                append whitespace " "
            }
        }
        $w delete $curr "$curr+${count}c"
        $w insert $curr $whitespace
        set oldcurr $curr
        set curr [$w index "$curr+1lines"]
        if {[$w compare $oldcurr == $curr]} {
            break
        }
    }
    $w tag add sel $first $last
    return
}


set gdm_window_placement_offset_x 0
set gdm_window_placement_offset_y 0

proc place_window_default2 {window parent} {
    global gdm_window_placement_offset_x
    global gdm_window_placement_offset_y

    if {![winfo exists $window]} {
        return
    }
    if {![winfo exists $parent]} {
        set parent .mw
    }

    incr gdm_window_placement_offset_x 25
    incr gdm_window_placement_offset_y 25
    if {$gdm_window_placement_offset_x > 150} {
        set gdm_window_placement_offset_x 25
        set gdm_window_placement_offset_y 25
    } elseif {$gdm_window_placement_offset_y > 150} {
        set gdm_window_placement_offset_x 25
        set gdm_window_placement_offset_y 25
    }

    set ptop [winfo toplevel $parent]

    set x [expr {[winfo rootx $ptop] + $gdm_window_placement_offset_x}]
    set y [expr {[winfo rooty $ptop] + $gdm_window_placement_offset_y}]

    set width [winfo width $window]
    set screenwidth [winfo screenwidth $window]
    if {$x + $width > $screenwidth - 20} {
        incr x [expr {($screenwidth - 20) - ($x + $width)}]
    }

    set height [winfo height $window]
    set screenheight [winfo screenheight $window]
    if {$y + $height > $screenheight - 45} {
        incr y [expr {($screenheight - 45) - ($y + $height)}]
    }

    after idle wm geometry $window "+$x+$y"
}


proc place_window_default {window {parent ""}} {
    if {$parent == ""} {
        set parent ".mw"
    }

    set sx [winfo screenwidth $window]
    set sy [winfo screenheight $window]
    incr sx 25
    incr sy 25
    wm geometry $window "+$sx+$sy"
    wm deiconify $window
    wm sizefrom $window program
    wm positionfrom $window program
    after idle place_window_default2 $window $parent
}


proc tkshellcmd {command} {
    if {[eof stdin]} {
        exit
    }
    gets stdin line
    if {$command != ""} {
        append command "\n"
    }
    append command $line
    if {[info complete $command]} {
        if {[catch {uplevel #0 $command} result]} {
            global errorInfo
            puts stderr $errorInfo
            flush stderr
        } else {
            puts stdout $result
        }
        set command ""
        puts -nonewline stdout "% "
        flush stdout
    } else {
        puts -nonewline stdout "> "
        flush stdout
    }
    fileevent stdin readable [list tkshellcmd $command] 
}


proc main {argc argv} {
    global sok treb_version treb_save_file
    global tcl_platform treb_fonts treb_lib_dir
    if {[lsearch -exact $argv "--tkshell"] != -1} {
        puts -nonewline stdout "% "
        fileevent stdin readable [list tkshellcmd ""] 
        return
    }

    set starttime [clock sec]

    Window show .mw
    if {[llength [info procs vTcl*]] < 20} {
        init_menus .mw
        standard_timer
    }

    /statbar 0 "Loading preferences..."
    if {![/load $treb_save_file]} {
        /bind add <KP_Enter>             {/dokey enter}
        /bind add <Return>               {/dokey enter}
        if {$tcl_platform(platform) == "macintosh" || $tcl_platform(os) == "Darwin"} {
            /bind add <Meta-Return>      {/dokey insert_enter}
        } else {
            /bind add <Control-Return>   {/dokey insert_enter}
            /bind add <Control-Next>     {/dokey scroll_line_dn}
            /bind add <Control-Prior>    {/dokey scroll_line_up}
        }
        /bind add <Shift-Next>           {/dokey scroll_line_dn}
        /bind add <Shift-Prior>          {/dokey scroll_line_up}
        /bind add <Next>                 {/dokey scroll_page_dn}
        /bind add <Prior>                {/dokey scroll_page_up}
        /bind add <Tab>                  {/dokey complete_word}
        if {$tcl_platform(platform) == "macintosh" || $tcl_platform(os) == "Darwin"} {
            /bind add <Meta-u>           {/dokey del_to_start}
            /bind add <Meta-Tab>         {/dokey socket_next}
            /bind add <Meta-Shift-Tab>   {/dokey socket_prev}
            /bind add <Meta-l>           {/dokey clear_screen}
            /bind add <Meta-Next>        {/dokey history_next}
            /bind add <Meta-Prior>       {/dokey history_prev}
            /bind add <Meta-Delete>      {/dokey delete_word}
        } else {
            /bind add <Control-u>            {/dokey del_to_start}
            /bind add <Control-Tab>          {/dokey socket_next}
            /bind add <Control-Shift-Tab>    {/dokey socket_prev}
            /bind add <Control-l>            {/dokey clear_screen}
            /bind add <Control-Key-n>        {/dokey history_next}
            /bind add <Control-Key-p>        {/dokey history_prev}
            /bind add <Control-Key-w>        {/dokey delete_word}
        }
    }

    if {![/style:exists normal]} {
        /style add normal -lmargin2 10 -font $treb_fonts(fixed)
    }
    if {![/style:exists hilite]} {
        /style add hilite -background yellow
    }
    if {![/style:exists error]} {
        /style add error -background black -foreground red  -font [concat $treb_fonts(sansserif) bold]
    }
    if {![/style:exists results]} {
        /style add results -background black -foreground green -font [concat $treb_fonts(sansserif) bold]
    }
    if {![/style:exists url]} {
        /style add url -foreground blue2 -font [concat $treb_fonts(sansserif) bold underline] -relief flat -borderwidth 0
        /style menu add url {Copy URL to clipboard} {/clipboard:copy %!}
        /style menu add url {Open URL...} {/web_view %!}
        /style menu add url {<<Click>>} {/web_view %!}
    }

    if {[/prefs:get show_qbuttons]} {
        /qbutton:showbar
    } else {
        /qbutton:hidebar
    }


    /statbar 0 "Running user defined startup commands..."
    catch [/prefs:get startup_script]

    #############################################
    # Add tags to show the default title screen #
    #############################################
    [/display] tag configure _wordtreb -font {Times 12 bold italic}
    [/display] tag configure _deftreb  -font {Times 12} -lmargin1 20 -lmargin2 40
    [/display] tag configure _bigtreb \
        -font {Helvetica 36 bold italic} -foreground blue4 -justify center
    [/display] tag configure _medtreb \
        -font {Helvetica 18 bold} -foreground green4 -justify center
    [/display] tag configure _smltreb \
        -font {Times 14 bold italic} -foreground black -justify center

    /echo -style _wordtreb {Treb'u*chet}
    /echo -style _deftreb -newline 0 { (Treb-oo-shay), n.}
    /echo -style _deftreb {1. A military engine used in the Middle Ages for throwing stones, etc. It acted by means of a great weight fastened to the short arm of a lever, which, being let fall, raised the end of the long arm with great velocity, hurling stones with much force.}
    /echo -style _deftreb {2. A MUCK client program by Fuzzball Software.}
    /echo
    /echo -style _medtreb "Fuzzball Software Presents:"
    /echo -style _bigtreb "Trebuchet Tk"
    /echo -style _medtreb "Version $treb_version"
    /echo -style _smltreb {"An excellent way to sling MUD."}
    /echo
    /echo

    /statbar 0 "Trebuchet is now ready to use."

    global dirty_preferences
    set dirty_preferences 0

    set warnver 8.030201
    set currver [format "%d.%02d%02d01" $tcl_platform(vermajor) $tcl_platform(verminor) $tcl_platform(patchlevel)]
    if {$currver < $warnver && [/prefs:get version_warning] < $warnver} {
        set mesg "You should use TCL/Tk 8.3.2, or a more recent interpreter to\n"
        append mesg "run TrebuchetTk. You can still run Trebuchet with an older 8.0\n"
        append mesg "interpreter, but you may experience slightly buggy behaviour,\n"
        append mesg "and you won't be able to use some features like SSL encryption.\n"
        append mesg "\n"
        append mesg "You can fetch and install the latest TCL/Tk interpreter from"
        set tclurl "http://www.belfry.com/fuzzball/trebuchet/faq.html#WhereGetIt"
        set mesg2 "This warning won't be shown again, once you save preferences."
        set base [toplevel .verwarn]
        label $base.icon -bitmap warning -foreground "#990"
        label $base.text -text $mesg -anchor sw -justify left -font $treb_fonts(sansserif)
        label $base.url -text $tclurl -anchor nw -justify left -foreground blue -font [concat $treb_fonts(sansserif) underline] -cursor hand2
        label $base.text2 -text $mesg2 -anchor nw -justify left -font $treb_fonts(sansserif)
        button $base.btn -text Okay -command "destroy $base"
        bind $base.url <ButtonPress-1> "
                $base.url config -foreground red
                /web_view [list $tclurl]
                after 250 $base.url config -foreground blue
            "
        pack $base.btn  -side bottom -anchor s -padx 10 -pady 10
        pack $base.icon  -side left -anchor w -padx 10 -pady 10
        pack $base.text  -side top -anchor w -expand 1 -fill x -padx 10 -pady 5
        pack $base.url   -side top -anchor nw -expand 1 -fill x -padx 10 -pady 0
        pack $base.text2 -side top -anchor nw -expand 1 -fill both -padx 10 -pady 5
        wm title $base "TCL Version [info tclversion]"
        tkwait window $base

        /prefs:set version_warning $warnver
    }

    global treb_revision
    if {[/prefs:get lastrun_version] < 124} {
        if {$tcl_platform(platform) == "unix" && $tcl_platform(os) != "Darwin"} {
            /style:resizeall 60
        }
    }
    if {[/prefs:get lastrun_version] != $treb_revision} {
        /prefs:set lastrun_version $treb_revision
    }

    if {[/prefs:get hide_splash]} {
        /tool:load
        main2 $argc $argv
    } else {
        Window show .
        /tool:load
        after 5000 "
            wm withdraw .
            destroy .splash
            image delete trebfling
            main2 [list $argc] [list $argv]
        "
    }
}

proc main2 {argc argv} {
    wm deiconify .mw
    update idletasks
    focus -force [/inbuf]

    if {[/prefs:get autoup_firsttime]} {
        set result [tk_messageBox -default yes -type yesno -icon question \
            -title "Network Update Check" -message "Would you like Trebuchet to automatically check for upgrades when it starts up in the future?\nThis dialog will not be shown again, once you save your preferences."]
        if {$result == "yes"} {
            /prefs:set autoupdate_check 1
        } else {
            /prefs:set autoupdate_check 0
        }
        /prefs:set autoup_firsttime 0
    }
    if {[/prefs:get autoupdate_check]} {
        /statbar 5 "Contacting Belfry.com to check for Network Updates."
        /web_upgrade -nowarns -timeout 10000
    }
    if {[/prefs:get startup_con_dlog]} {
        /selector -contentscript {/world names} -register {/world} \
            -title "Connect to world" \
            -caption {Select a world to connect to} \
            -selectbutton "Connect" -selectscript {/connect %!} \
            -editbutton "Edit" -editscript {/world:edit %!} -editpersist
    }
}

proc Window {args} {
    global vTcl
    set cmd [lindex $args 0]
    set name [lindex $args 1]
    set newname [lindex $args 2]
    set rest [lrange $args 3 end]
    if {$name == "" || $cmd == ""} {return}
    if {$newname == ""} {
        set newname $name
    }
    set exists [winfo exists $newname]
    switch $cmd {
        show {
            if {$exists == "1" && $name != "."} {wm deiconify $name; return}
            if {[info procs vTclWindow(pre)$name] != ""} {
                eval "vTclWindow(pre)$name $newname $rest"
            }
            if {[info procs vTclWindow$name] != ""} {
                eval "vTclWindow$name $newname $rest"
            }
            if {[info procs vTclWindow(post)$name] != ""} {
                eval "vTclWindow(post)$name $newname $rest"
            }
        }
        hide    { if {$exists} {wm withdraw $newname; return} }
        iconify { if {$exists} {wm iconify $newname; return} }
        destroy { if {$exists} {destroy $newname; return} }
    }
}

#################################
# VTCL GENERATED GUI PROCEDURES
#

proc vTclWindow. {base} {
    global treb_version treb_lib_dir tcl_platform

    if {$base == ""} {
        set base .
    }
    ###################
    # CREATING WIDGETS
    ###################
    wm withdraw $base
    wm resizable $base 0 0
    wm title $base "Trebuchet Tk"
    wm protocol $base WM_DELETE_WINDOW /quit

    . config -relief solid -borderwidth 0
    frame .splash -relief raised -borderwidth 4
    frame .splash.border -relief flat -borderwidth 4
    image create photo trebfling -file [file join $treb_lib_dir images trbsplsh.gif]
    label .splash.img -image trebfling -relief sunken -borderwidth 1
    pack .splash
    pack .splash.border
    pack .splash.img -in .splash.border

    set newx [expr {([winfo screenwidth .] / 2) - ([image width trebfling] / 2) - 5}]
    set newy [expr {([winfo screenheight .] / 2) - ([image height trebfling] / 2) - 25}]

    wm overrideredirect $base 1
    wm deiconify $base
    wm geometry $base +$newx+$newy
    update
    ###################
    # SETTING GEOMETRY
    ###################
}

proc vTclWindow.mw {base} {
    global argv0 env treb_fonts
    global tcl_platform
    if {$base == ""} {
        set base .mw
    }
    if {[winfo exists $base]} {
        wm deiconify $base; return
    }
    ###################
    # CREATING WIDGETS
    ###################
    toplevel $base -class Toplevel -borderwidth 0
    wm withdraw $base
    wm focusmodel $base passive
    wm geometry $base 670x588
    wm minsize $base 160 240
    wm overrideredirect $base 0
    wm resizable $base 1 1
    wm title $base "Trebuchet Tk"
    wm iconname $base "TrebTk"
    wm group $base $base
    wm command $base $argv0
    wm protocol $base WM_DELETE_WINDOW /quit
    frame $base.con \
        -height 100 -width 200 -borderwidth 0
    frame $base.con.top -borderwidth 0
    frame $base.con.top.con \
        -relief flat -borderwidth 0
    buttonbar:new $base.con.top.con.qbuttons "QuickButtons toolbar" -newcmd {
        /newdlog QuickButton /qbutton
    } -editcmd {
        /qbutton:edit %!
    } -editbbarcmd {
        /qbutton:edit
    }
    /qbutton:hidebar
    frame $base.con.top.con.disp \
        -borderwidth 0 -relief flat
    frame $base.con.top.con.worlds \
        -borderwidth 2 -relief flat -height 25
    frame $base.con.bot -borderwidth 0
    frame $base.con.bot.con \
        -height 30 -width 30 -relief flat -borderwidth 0
    if {$tcl_platform(os) == "Darwin"} {
        frame $base.con.bot.con.in \
            -height 30 -width 30 -relief flat -borderwidth 0
    } else {
        frame $base.con.bot.con.in \
            -height 30 -width 30 -relief sunken -borderwidth 2
    }
    set sbarbw 0
    if {$tcl_platform(platform) == "unix"} {
        if {$tcl_platform(os) != "Darwin"} {
            set sbarbw 1
        }
    }
    scrollbar $base.con.bot.con.in.scroll \
        -command {.mw.con.bot.con.in.inbuf yview} -orient vert \
        -borderwidth $sbarbw -relief raised
    text $base.con.bot.con.in.inbuf \
        -font $treb_fonts(fixed) -height 1 -width 8 -relief flat -borderwidth 0 \
        -yscrollcommand {.mw.con.bot.con.in.scroll set} \
        -highlightthickness 0 -insertwidth 2
    textPopup:new $base.con.bot.con.in.inbuf
    catch {
        global tcl_platform
        switch -glob -- $tcl_platform(os) {
            Win* {
                bind $base.con.bot.con.in.inbuf <MouseWheel> {
                    if {%D > 0} {
                        /dokey scroll_line_up
                        /dokey scroll_line_up
                        /dokey scroll_line_up
                        /dokey scroll_line_up
                    } else {
                        /dokey scroll_line_dn
                        /dokey scroll_line_dn
                        /dokey scroll_line_dn
                        /dokey scroll_line_dn
                    }
                }
            }
            default {
                bind $base.con.bot.con.in.inbuf <Button-4> {
                    /dokey scroll_line_up
                    /dokey scroll_line_up
                    /dokey scroll_line_up
                    /dokey scroll_line_up
                }
                bind $base.con.bot.con.in.inbuf <Button-5> {
                    /dokey scroll_line_dn
                    /dokey scroll_line_dn
                    /dokey scroll_line_dn
                    /dokey scroll_line_dn
                }
            }
        }
    }
    bind $base.con.bot.con.in.inbuf <Key> {+
        after 10 update_charcount
    }
    bind $base.con.bot.con.in.inbuf <Button-1> {+
        after 10 update_charcount
    }
    bind $base.con.bot.con.in.inbuf <<Cut>> {+
        after 10 update_charcount
    }
    bind $base.con.bot.con.in.inbuf <<Paste>> {+
        after 10 update_charcount
    }
    if {$tcl_platform(os) != "Darwin"} {
        set statrel "sunken"
    } else {
        set statrel "flat"
    }
    frame $base.statusbar \
        -height 30 -width 30 -relief flat -borderwidth 0
    button $base.statusbar.histbtn \
        -width 11 -image [gdm:Bitmap get incr] -padx 0 -pady 0 -command {
            /results -title "Status bar message history" -showend [/statbar:list]
        }
    label $base.statusbar.note \
        -text {} -relief flat -borderwidth 1 -anchor w
    label $base.statusbar.secure \
        -image ssl_icon_insecure -relief $statrel -borderwidth 1 -anchor center
    label $base.statusbar.log \
        -text {} -relief $statrel -borderwidth 1 -width 5 -anchor center
    frame $base.statusbar.pos \
        -relief $statrel -borderwidth 1
    label $base.statusbar.pos.line \
        -text {} -relief flat -borderwidth 0 -width 3 -anchor w
    label $base.statusbar.pos.char \
        -text {} -relief flat -borderwidth 0 -width 4 -anchor w

    set statbarimg [gdm:Bitmap get statbar]
    label $base.statusbar.corner \
        -image $statbarimg -relief flat -borderwidth 0 -padx 0 -pady 0

    if {$tcl_platform(os) != "Darwin"} {
        switch -exact -- $tcl_platform(platform) {
            windows {$base.statusbar.corner conf -cursor fleur}
            unix    {$base.statusbar.corner conf -cursor bottom_right_corner}
        }

        set treb_window_resizing_offset {0 0}
        set treb_window_resizing_pos +0+0

        bind $base.statusbar.corner <ButtonPress-1> {movebox:btnpress %W %X %Y}
        proc movebox:btnpress {wname mousex mousey} {
            upvar #0 treb_window_resizing_offset offset
            upvar #0 treb_window_resizing_pos pos
            set top [winfo toplevel $wname]
            set x [winfo rootx $top]
            set y [winfo rooty $top]
            set xoff [expr {$mousex - $x - [winfo width $top]}]
            set yoff [expr {$mousey - $y - [winfo height $top]}]
            set offset [list $xoff $yoff]
            global tcl_platform
            if {$tcl_platform(platform) == "unix"} {
                set topgeom [wm geom $top]
                set topy [lindex [split $topgeom "+"] 2]
                set topydelt [expr {$topy - [winfo rooty $top]}]
                incr y $topydelt
            }
            set pos +$x+$y
        }

        bind $base.statusbar.corner <B1-Motion> {movebox:motion %W %X %Y}
        proc movebox:motion {wname mousex mousey} {
            upvar #0 treb_window_resizing_offset offset
            upvar #0 treb_window_resizing_pos pos
            set top [winfo toplevel $wname]
            set x [winfo rootx $top]
            set y [winfo rooty $top]
            set w [expr {$mousex - [lindex $offset 0] - $x}]
            set h [expr {$mousey - [lindex $offset 1] - $y}]
            set minsize [wm minsize $top]
            set minw [lindex $minsize 0]
            set minh [lindex $minsize 1]
            if {$w < $minw} { set w $minw }
            if {$h < $minh} { set h $minh }
            wm geometry $top ${w}x${h}${pos}
        }
    }

    bind $base.statusbar.secure <1> {
        /tls:showcert
    }

    frame $base.con.thumb \
        -borderwidth 2 -cursor sb_v_double_arrow \
        -relief raised 
    proc pane:movethumb {root y} {
        set th [expr {2 * [winfo height $root.thumb] + 4}]
        set height [expr {[winfo height $root] - $th}]
        set fh [font metrics [/inbuf cget -font] -linespace]
        set y [expr {$height - $fh * int(($height - $y + $th) / $fh)}]
        set val [expr {($y - [winfo rooty $root]) / (($height + 0.0) + $th)}]

        if { $val >= 0 && $val <= 1.0 } {
            place $root.thumb -rely $val -y 2
        }
    }
    proc pane:setthumb {root y} {
        set th [expr {2 * [winfo height $root.thumb] + 4}]
        set height [expr {[winfo height $root] - $th}]
        set fh [font metrics [/inbuf cget -font] -linespace]
        set y [expr {$height - $fh * int(($height - $y + $th) / $fh)}]
        set val [expr {($y - [winfo rooty $root]) / (($height + 0.0) + $th)}]

        if { $val >= 0 && $val <= 1.0 } {
            place $root.top -relheight $val
            place $root.thumb -rely $val -y 2
            place $root.bot -relheight [ expr {1.0 - $val}]
            update idletasks
        }
    }
    bind $base.con.thumb <B1-Motion> {pane:movethumb .mw.con %Y}
    bind $base.con.thumb <ButtonRelease-1> {pane:setthumb .mw.con %Y}

    ###################
    # SETTING GEOMETRY
    ###################
    pack $base.statusbar \
        -side bottom -anchor s -fill x
    pack $base.statusbar.corner \
        -anchor se -expand 0 -fill none -side right -padx 0
    pack $base.statusbar.secure \
        -anchor se -expand 0 -fill none -side right -padx 1
    pack $base.statusbar.log \
        -anchor se -expand 0 -fill none -side right -padx 1
    pack $base.statusbar.pos \
        -anchor se -expand 0 -fill none -side right -padx 1
    pack $base.statusbar.pos.char \
        -anchor se -expand 0 -fill none -side right -padx 1
    pack $base.statusbar.pos.line \
        -anchor se -expand 0 -fill none -side right -padx 1
    pack $base.statusbar.histbtn \
        -anchor se -expand 0 -fill y -side right -padx 1
    pack $base.statusbar.note \
        -anchor nw -expand 1 -fill x -side left -padx 1
    pack $base.con \
        -in .mw -anchor center -expand 1 -fill both -side top 
    place $base.con.top \
        -x 0 -y 0 -relwidth 1 -height -1 -relheight 0.87 -anchor nw \
        -bordermode ignore 
    pack $base.con.top.con \
        -anchor center -expand 1 -fill both -side top 
    pack $base.con.top.con.worlds \
        -anchor s -expand 0 -fill x -side bottom
    pack .mw.con.top.con.qbuttons \
        -anchor n -expand 0 -fill x -side top -padx 1
    pack $base.con.top.con.disp \
        -anchor center -side bottom -expand 1 -fill both -padx 0
    place $base.con.bot \
        -x 0 -y 0 -rely 1 -relwidth 1 -height -1 -relheight 0.13 \
        -anchor sw -bordermode ignore 
    pack $base.con.bot.con \
        -anchor center -expand 1 -fill both -side top 
    grid columnconf $base.con.bot.con 0 -weight 1
    grid rowconf $base.con.bot.con 0 -weight 1
    grid $base.con.bot.con.in \
        -column 0 -row 0 -columnspan 1 -rowspan 1 -sticky nesw -pady 2 -padx 0
    grid columnconf $base.con.bot.con.in 0 -weight 1
    grid rowconf $base.con.bot.con.in 0 -weight 1
    grid $base.con.bot.con.in.scroll \
        -column 1 -row 0 -columnspan 1 -rowspan 1 -sticky ns 
    grid $base.con.bot.con.in.inbuf \
        -column 0 -row 0 -columnspan 1 -rowspan 1 -sticky nesw
    place $base.con.thumb \
        -x 0 -relx 0 -y 2 -rely 0.87 -width 0 -relwidth 1 -height 4 \
        -anchor w -bordermode ignore 
    display:mkdisplay $base.con.top.con.disp.backdrop
    pack $base.con.top.con.disp.backdrop \
        -anchor center -expand 1 -fill both -side top
    after 100 "bind $base.con.bot.con.in.inbuf <Configure> {+after 10 show_mainwin_size}"
}

proc show_mainwin_size {} {
    set disp [/display]
    set dwidth [winfo width $disp]
    set dheight [winfo height $disp]
    incr dwidth -2
    incr dheight -2
    set font [$disp cget -font]
    set fwidth [font measure $font "0"]
    set fheight [font metrics $font -linespace]
    set height [expr {int($dheight / $fheight)}]
    set width [expr {int($dwidth / $fwidth)}]
    /statbar -nolog 3 "${width}x${height}"
}

main $argc $argv


