tcl::OptProc fakebutton:new {
    {wname            {}    {Widget name to create.}}
    {-text            {}    {Text to put on button.}}
    {-image           {}    {Image to put on button.}}
    {-width           0     {Width of button.}}
    {-height          0     {Height of button.}}
    {-command         {}    {Command to execute when button is pressed.}}
    {-borderwidth     2     {Width of 3D bezel around button.}}
    {-relief  -choice {raised flat} {Relief of unpressed 3D bezel.}}
} {
    global treb_clickedin treb_fonts
    set treb_clickedin {}
    if {[catch {
        frame $wname -class FakeBtn -relief $relief -borderwidth $borderwidth -highlightthickness 0 -padx 3
    }]} {
        frame $wname -class FakeBtn -relief $relief -borderwidth $borderwidth -highlightthickness 0
    }
    frame $wname.ff -highlightthickness 1 -takefocus 1
    bindtags $wname.ff [list $wname $wname.ff [winfo toplevel $wname] all]
    pack $wname.ff -expand 1 -fill both
    if {$width > 0} { $wname config -width $width }
    if {$height > 0} { $wname config -height $height }
    if {$image != {}} {
        label $wname.ff.img -image $image -borderwidth 0 -padx 0 -highlightthickness 0
        bindtags $wname.ff.img [list $wname $wname.ff.img [winfo toplevel $wname] all]
        pack $wname.ff.img -expand 0 -fill none -side left
    }
    if {$text != {}} {
        label $wname.ff.lbl -text $text -borderwidth 0 -padx 2 -highlightthickness 0 -font $treb_fonts(bbar)
        if {$image != {}} {
            $wname.ff.lbl config -anchor w
        }
        bindtags $wname.ff.lbl [list $wname $wname.ff.lbl Label [winfo toplevel $wname] all]
        pack $wname.ff.lbl -expand 1 -fill both -side left
    }

    # Button is Up.
    set uprel $relief
    set upbg [$wname cget -background]

    # Button is Down.
    set dnrel "sunken"
    set dnbg $upbg

    # Hovering over Button.
    set actrel "raised"
    set actbg $upbg

    bind $wname <Enter> "if {\$treb_clickedin == \"$wname\"} {$wname config -relief $dnrel -background $dnbg} else {$wname config -relief $actrel -background $actbg}; update idletasks"
    bind $wname <Leave> "$wname config -relief $uprel -background $upbg ; update idletasks"
    bind $wname <ButtonPress-1> "$wname config -relief $dnrel -background $dnbg ; set treb_clickedin $wname; update idletasks"
    bind $wname <ButtonRelease-1> "if {\$treb_clickedin == \"$wname\" && \[$wname cget -relief\] == \"$dnrel\"} {eval [list $command]} ; $wname config -relief $uprel -background $upbg ; set treb_clickedin {}; update idletasks"
    bind $wname <Key-space> "$wname config -relief $dnrel -background $dnbg ; update idletasks ; eval [list $command] ; after 200 \"$wname config -relief $uprel -background $upbg ; update idletasks\" ; set treb_clickedin {}"
}

proc fakebutton:width {wname} {
    set width 0
    if {[winfo exists $wname.ff.lbl]} {
        set font [$wname.ff.lbl cget -font]
        set text [$wname.ff.lbl cget -text]
        incr width [font measure $font $text]
        incr width [expr {2 * [$wname.ff.lbl cget -padx]}]
        incr width [expr {2 * [$wname.ff.lbl cget -borderwidth]}]
        incr width [expr {2 * [$wname.ff.lbl cget -highlightthickness]}]
    }
    if {[winfo exists $wname.ff.img]} {
        set img [$wname.ff.lbl cget -image]
        incr width [image width $img]
        incr width [expr {2 * [$wname.ff.img cget -padx]}]
        incr width [expr {2 * [$wname.ff.img cget -borderwidth]}]
        incr width [expr {2 * [$wname.ff.img cget -highlightthickness]}]
    }
    incr width [expr {2 * [$wname.ff cget -borderwidth]}]
    incr width [expr {2 * [$wname.ff cget -highlightthickness]}]
    incr width [expr {2 * [$wname cget -borderwidth]}]
    incr width [expr {2 * [$wname cget -highlightthickness]}]
    catch {
        incr width [expr {2 * [$wname cget -padx]}]
    }
    return $width
}


tcl::OptProc buttonbar:new {
    {base       {}      {The widget name of the buttonbar to create.}}
    {title      {}      {The title of the buttonbar to create, for menus.}}
    {-newcmd    {}      {The command to run when context "new button" is selected.}}
    {-editcmd   {}      {The command to run when context "edit button" is selected.}}
    {-editbbarcmd {}    {The command to run when context "edit buttonbar" is selected.}}
} {
    global treb_buttonbar tcl_platform
    set relief "flat"
    if {$tcl_platform(os) != "Darwin"} {
        set relief "sunken"
    }
    frame $base -borderwidth 2 -relief $relief -height 25
    if {![info exists treb_buttonbar(bars)]} {
        set treb_buttonbar(bars) {}
    }
    lappend treb_buttonbar(bars) $base
    set treb_buttonbar(title,$base) $title
    set treb_buttonbar(order,$base) {}
    set treb_buttonbar(visible,$base) 1
    set treb_buttonbar(newcmd,$base) $newcmd
    set treb_buttonbar(editcmd,$base) $editcmd
    set treb_buttonbar(editbbarcmd,$base) $editbbarcmd
    bind $base <Configure> "buttonbar:refresh $base 60"
    bind $base <<ContextMenu>> "buttonbar:contextmenu $base {}"
}

proc buttonbar:delete {base id} {
    global treb_buttonbar
    if {![info exists treb_buttonbar(order,$base)]} {
        return
    }
    set pos [lsearch -exact $treb_buttonbar(order,$base) $base.$id]
    set treb_buttonbar(order,$base) [lreplace $treb_buttonbar(order,$base) $pos $pos]
    set treb_buttonbar(cols,$base) 0
    destroy $base.$id
    buttonbar:refresh $base 60
    return
}

proc buttonbar:show {base {state 1}} {
    global treb_buttonbar
    if {$state != 0} {
        if {[winfo ismapped $base]} {
            set mgr [winfo manager $base]
            if {$mgr != {}} {
                set treb_buttonbar(mgrinfo,$base) "$mgr $base [$mgr info $base]"
            }
        } else {
            if {[info exists treb_buttonbar(mgrinfo,$base)]} {
                eval $treb_buttonbar(mgrinfo,$base)
            }
        }
    } else {
        set mgr [winfo manager $base]
        if {$mgr != {}} {
            set treb_buttonbar(mgrinfo,$base) "$mgr $base [$mgr info $base]"
            eval "$mgr forget $base"
        }
    }
}

proc buttonbar:contextmenu {base {data {}}} {
    global treb_buttonbar
    set treb_buttonbar(visible,$base) [winfo ismapped $base]
    set toplev [winfo toplevel $base]
    set menu $toplev.bbarcontextmenu
    if {[winfo exists $menu]} {
        destroy $menu
    }

    menu $menu -tearoff false

    if {$treb_buttonbar(newcmd,$base) != {}} {
        $menu add command -label "New button..." -underline 0 -command "eval \$treb_buttonbar(newcmd,$base)"
    }

    set state "disabled"
    if {$data != {} && $treb_buttonbar(editcmd,$base) != {}} {
        set state "normal"
    }
    $menu add command -label "Edit button..." -underline 0 -state $state -command "eval \[/line_subst \$treb_buttonbar(editcmd,$base) [list $data]\]"

    if {$treb_buttonbar(editbbarcmd,$base) != {}} {
        $menu add command -label "Edit buttonbar..." -underline 0 -command "eval \$treb_buttonbar(editbbarcmd,$base)"
    }

    $menu add separator
    foreach bbar $treb_buttonbar(bars) {
        $menu add checkbutton -label $treb_buttonbar(title,$bbar) \
            -variable treb_buttonbar(visible,$bbar) -command "
                buttonbar:show $base \$treb_buttonbar(visible,$bbar)
            "
    }
    tk_popup $menu [winfo pointerx $base] [winfo pointery $base]
}

tcl::OptProc buttonbar:add {
    {base       {}      {The buttonbar widget to add a button to.}}
    {?pos?      {end}   {The position to insert the button before.}}
    {-image     {}      {The Icon to display in the button.}}
    {-text      {}      {The text to display in the button.}}
    {-command   {}      {The command to execute when the button is pressed.}}
    {-data      {}      {Instance specific data.}}
} {
    global treb_buttonbar tcl_platform
    set i 0
    while {[winfo exists $base.$i]} {
        incr i
    }
    set relief "flat"
    set bwidth 1
    if {$tcl_platform(os) != "Darwin"} {
        set relief "raised"
        set bwidth 2
    }
    fakebutton:new $base.$i -text $text -image $image -command $command \
        -borderwidth $bwidth -relief $relief
    if {![info exists treb_buttonbar(order,$base)]} {
        set treb_buttonbar(order,$base) {}
    }
    if {$pos == "end"} {
        set pos [llength $treb_buttonbar(order,$base)]
    }
    set treb_buttonbar(order,$base) [linsert $treb_buttonbar(order,$base) $pos $base.$i]
    set treb_buttonbar(widths,$base.$i) [fakebutton:width $base.$i]
    set treb_buttonbar(data,$base.$i) $data
    set treb_buttonbar(cols,$base) 0
    buttonbar:refresh $base 60
    bind $base.$i <<ContextMenu>> "buttonbar:contextmenu $base [list $data]"
    return $i
}

proc buttonbar:calccols {base minwidth cols} {
    if {$cols <= 1} { return 1 }
    global treb_buttonbar
    for {set col 0} {$col < $cols} {incr col} {
        set treb_buttonbar(colwidths,$base,$col) 0
    }
    set col 0
    set totalwidth 0
    set winwidth [winfo width $base]
    foreach child $treb_buttonbar(order,$base) {
        set childwidth $treb_buttonbar(widths,$child)
        if {$childwidth < $minwidth} {
            set childwidth $minwidth
        }
        if {$childwidth > $treb_buttonbar(colwidths,$base,$col)} {
            set treb_buttonbar(colwidths,$base,$col) $childwidth
        }
        incr totalwidth $treb_buttonbar(colwidths,$base,$col)
        if {$totalwidth >= $winwidth} {
            return [buttonbar:calccols $base $minwidth $col]
        }
        incr col
        if {$col >= $cols} {
            set col 0
            set totalwidth 0
        }
    }
    set totalwidth 0
    for {set col 0} {$col < $cols} {incr col} {
        incr totalwidth $treb_buttonbar(colwidths,$base,$col)
        if {$totalwidth >= $winwidth} {
            return [buttonbar:calccols $base $minwidth $col]
        }
    }
    return $cols
}

proc buttonbar:refresh {base minwidth} {
    global treb_buttonbar_refreshing
    if {[info exists treb_buttonbar_refreshing]} return
    set treb_buttonbar_refreshing 1
    global treb_buttonbar
    set cnt [llength $treb_buttonbar(order,$base)]
    set cols [buttonbar:calccols $base $minwidth $cnt]
    if {![info exists treb_buttonbar(order,$base)]} {
        set treb_buttonbar(order,$base) {}
    }
    if {[info exists treb_buttonbar(cols,$base)]} {
        if {$cols == $treb_buttonbar(cols,$base)} {
            unset treb_buttonbar_refreshing
            return
        }
    }
    set treb_buttonbar(cols,$base) $cols
    set row 0
    set col 0
    foreach child $treb_buttonbar(order,$base) {
        if {$col >= $cols} {
            set col 0
            incr row
        }
        grid columnconf $base $col -weight 0
        grid rowconf $base $row -weight 0
        grid $child -row $row -column $col -sticky nsew
        incr col
    }
    grid columnconf $base $cols -weight 1
    update idletasks
    unset treb_buttonbar_refreshing
}

proc buttonbar:clear {base} {
    foreach btn [winfo children $base] {
        destroy $btn
    }
}


