###########################
##  AWNS Ping Interface  ##
############################################################################
##
##
############################################################################

global PingInfo


proc awns_ping_init {world vers} {
    set pkg "dns-com-awns-ping"
    /menu:add "Local|Ping..." -command "/ping" -statecommand "ping:state"
}


proc awns_ping {world data} {
    set pkg "dns-com-awns-ping"

    array set arr $data
    setfrom id arr(id) {}

    /mcp_send $pkg "reply" -world $world id $id

    return
}


proc /ping {{world ""}} {
    global PingInfo

    set pkg "dns-com-awns-ping"
    if {$world == ""} {
        set world [/socket:foreground]
    }
    if {![/world:exists $world]} {
        return
    }
    if {[mcp_remote_pkg_supported $world $pkg] == ""} {
        return
    }

    while {1} {
        set    id [format "%04X" [expr {int(0x10000 * rand())}]]
        append id [format "%04X" [expr {int(0x10000 * rand())}]]
        if {![info exists PingInfo(starttime,$id)]} {
            break
        }
    }

    set PingInfo(starttime,$id) [/clock_clicks]

    /mcp_send $pkg "" -world $world id $id
}


proc awns_ping_reply {world data} {
    global PingInfo clicks_per_second

    set pkg "dns-com-awns-ping"

    array set arr $data
    setfrom id arr(id) {}

    if {[info exists PingInfo(starttime,$id)]} {
        set latency [expr {([/clock_clicks] - $PingInfo(starttime,$id)) * 1000 / $clicks_per_second}]
        unset PingInfo(starttime,$id)
        tk_messageBox -type ok -title "Ping $world" -message "Response time: $latency ms."
    }
    return
}

proc ping:state {} {
    set world [/socket:foreground]
    set pingpkg "dns-com-awns-ping"
    if {$world != ""} {
        if {[/socket:get state $world] != "Disconnected"} {
            if {[mcp_remote_pkg_supported $world $pingpkg] != ""} {
                return "normal"
            }
        }
    }
    return "disabled"
}


mcp_register_pkg "dns-com-awns-ping" 1.0 1.0 awns_ping_init
mcp_register_handler "dns-com-awns-ping" "" awns_ping
mcp_register_handler "dns-com-awns-ping" "reply" awns_ping_reply



proc awns_status {world data} {
    array set arr $data
    setfrom txt arr(text) {}

    /statbar 10 "$world - $txt"

    return
}


mcp_register_pkg "dns-com-awns-status" 1.0 1.0
mcp_register_handler "dns-com-awns-status" "" awns_status



tcl::OptProc /fbhelp {
    {-world    {}  "The world to get help from."}
    {-type     {}  "The type of help to get. (help, man, mpi, news)"}
    {topic     {}  "The topic to get help on."}
} {
    set pkg "org-fuzzball-help"

    if {$world == {}} {
        set world [/socket:current]
    }
    /mcp_send $pkg "request" -world $world type man topic $topic
}


proc fb_help_entry {world data} {
    set pkg "org-fuzzball-help"

    array set arr $data
    setfrom topic arr(topic) {}
    setfrom textval arr(text) {}

    if {[string first "-" $textval] == 0} {
        set textval "\n$textval"
    }
    /results -world $world -title "FuzzBall Help - $topic" -- $textval
    return
}

mcp_register_pkg "org-fuzzball-help" 1.0 1.0
mcp_register_handler "org-fuzzball-help" "entry" fb_help_entry
mcp_register_handler "org-fuzzball-help" "error" fb_help_entry



global LocalMenuInfo

proc fb_localmenu_enabled {name} {
    global LocalMenuInfo
    set world [/socket:foreground]
    regsub -all -nocase {[^A-Z_0-9|]} [string tolower $name] {} iname
    if {[info exists LocalMenuInfo(menuitem-$world-$iname)]} {
        return "normal"
    } else {
        return "disabled"
    }
}

proc fb_localmenu_report {world label id} {
    set pkg "org-fuzzball-localmenu"
    /mcp_send $pkg "selected" -world $world label $label id $id
}

proc fb_localmenu_add {world data} {
    global LocalMenuInfo
    set pkg "org-fuzzball-localmenu"

    array set arr $data
    setfrom itemtype  arr(type)  {}
    setfrom itemlabel arr(label) {}
    setfrom id        arr(id)    {}

    /menu:add "&Local|$itemlabel" -command "fb_localmenu_report [list $world] [list $itemlabel] [list $id]" -statecommand "fb_localmenu_enabled [list $itemlabel]"
    regsub -all -nocase {[^A-Z_0-9|]} [string tolower $itemlabel] {} iname
    unset LocalMenuInfo(menuitem-$world-$iname)
    return
}

proc fb_localmenu_del {world data} {
    global LocalMenuInfo
    set pkg "org-fuzzball-localmenu"

    array set arr $data
    setfrom itemlabel arr(label) {}
    setfrom id        arr(id)    {}

    /menu:del "&Local|$itemlabel"
    regsub -all -nocase {[^A-Z_0-9|]} [string tolower $itemlabel] {} iname
    set LocalMenuInfo(menuitem-$world-$iname) 1
    return
}

mcp_register_pkg "org-fuzzball-localmenu" 1.0 1.0
mcp_register_handler "org-fuzzball-localmenu" "add" fb_localmenu_add
mcp_register_handler "org-fuzzball-localmenu" "del" fb_localmenu_del

