proc send_dlog:do_quote {wname parent} {
    set world   [$wname.world get]
    set prepend [$wname.prependentry get]
    set append  [$wname.appendentry get]
    set file    [$wname.fileentry get]
    set delay   [$wname.delay get]

    destroy $wname
    textdlog:refocus $parent
    /prefs:set last_sent_file $file
    /prefs:set last_sent_prepend $prepend
    /prefs:set last_sent_append $append
    /prefs:set last_sent_delay $delay
    /quote -world $world -prefix $prepend -suffix $append -file $file -delay $delay
}

tcl::OptProc /send_dlog {
    {-wname		{}					"Toplevel window widget name"}
    {-parent	{}					"Widget to focus on after exiting"}
    {-title		{Send file to...}	"Toplevel window title"}
    {-world		{}					"The world to send the data to."}
    {-file		{}					"The file to quote to the given world."}
    {-prepend	{}					"String to prepend to each line."}
    {-append	{}					"String to append to each line."}
    {-command	{}					"A command to perform on each line."}
    {-delay		-float -1.0			"The time between lines sent, in seconds."}
} {
    if {$parent == {}} {
        set parent [focus]
        if {$parent == {}} {
            set parent [/inbuf]
        }
    }
    if {![winfo exists $parent]} {
        set parent [/inbuf]
    }
    if {$wname == {}} {
        global SendDlogNumber
        if {![info exists SendDlogNumber]} {
            set SendDlogNumber 0
        }
        incr SendDlogNumber
        set wname $wname.senddlog$SendDlogNumber
    }

    if {[winfo exists $wname]} {
        wm deiconify $wname
        focus $wname.world.entry
    } else {
        ###################
        # CREATING WIDGETS
        ###################
        toplevel $wname
        place_window_default $wname $parent
        wm resizable $wname 0 0
        wm title $wname $title
        label $wname.filelbl \
            -anchor w -borderwidth 1 -text {File} -underline 1 
        entry $wname.fileentry -width 50
        button $wname.filebrowse -text Browse... -underline 0 -command "
            set lastsent \[/prefs:get last_sent_file\]
            set file \[tk_getOpenFile -title {Select a file to send} \
                        -initialdir \[file dirname \$lastsent\] \
                        -initialfile \[file tail \$lastsent\] \
                        -parent $wname.fileentry -filetypes \
                        {{{Text Files} .txt TEXT} {{Log Files} .log TEXT} {{MUF Files} .muf TEXT} {{All Files} * }}\]
            unset lastsent
            if {\$file != {}} {
                $wname.fileentry delete 0 end
                $wname.fileentry insert end \$file
            }
        "
        label $wname.worldlbl -anchor w -borderwidth 1 -text {World} -underline 0 
        combobox $wname.world -editable 0
        label $wname.delaylbl -text {Seconds delay per line} -underline 8 -anchor w
        spinner $wname.delay -min 0.0 -max 9999.9 -val 0.0 -step 0.1 \
            -width 6

        label $wname.prependlbl \
            -anchor w -borderwidth 1 -text {Prepend to each line} -underline 1 
        entry $wname.prependentry
        label $wname.appendlbl \
            -borderwidth 1 -text {Append to each line} -underline 0 
        entry $wname.appendentry
        frame $wname.buttons \
            -borderwidth 2 -height 30 -width 125 
        button $wname.buttons.send \
            -text Send -underline 0 -width 6  -command "send_dlog:do_quote $wname $parent"
        button $wname.buttons.cancel \
            -text Cancel -width 6 -command "destroy $wname ; textdlog:refocus $parent"
        ###################
        # SETTING GEOMETRY
        ###################
        grid columnconf $wname  0 -minsize 10
        grid columnconf $wname  2 -minsize 10
        grid columnconf $wname  4 -minsize 10
        grid columnconf $wname  5 -weight 1
        grid columnconf $wname  6 -minsize 10
        grid columnconf $wname  8 -minsize 10
        grid columnconf $wname 10 -minsize 10

        grid rowconf $wname  0 -minsize 10
        grid rowconf $wname  2 -minsize 10
        grid rowconf $wname  4 -minsize 15
        grid rowconf $wname  6 -minsize 10
        grid rowconf $wname  8 -minsize 10
        grid rowconf $wname 10 -minsize 10

        grid $wname.filelbl      -column 1 -row  1 -sticky w 
        grid $wname.fileentry    -column 3 -row  1 -sticky ew   -columnspan 5
        grid $wname.filebrowse   -column 9 -row  1 -sticky nsew
        grid $wname.worldlbl     -column 1 -row  3 -sticky w
        grid $wname.world        -column 3 -row  3 -sticky nsew -columnspan 3
        grid $wname.delaylbl     -column 7 -row  3 -sticky w
        grid $wname.delay        -column 9 -row  3 -sticky w

        grid $wname.prependlbl   -column 1 -row  5 -sticky w    -columnspan 3
        grid $wname.prependentry -column 5 -row  5 -sticky nsew -columnspan 5
        grid $wname.appendlbl    -column 1 -row  7 -sticky w    -columnspan 3
        grid $wname.appendentry  -column 5 -row  7 -sticky nsew -columnspan 5
        grid $wname.buttons      -column 1 -row  9 -sticky nsew -columnspan 9

        grid columnconf $wname.buttons 0 -weight 1
        grid columnconf $wname.buttons 2 -minsize 10
        grid $wname.buttons.cancel -column 1 -row 1 -sticky nsew
        grid $wname.buttons.send   -column 3 -row 1 -sticky nesw 
        bind $wname <Key-Escape> "$wname.buttons.cancel invoke"
        bind $wname <Alt-Key-s> "$wname.buttons.send invoke"
    }

    wm title $wname $title

    $wname.world entrydelete 0 end
    foreach wrld [/socket:connectednames] {
        $wname.world entryinsert end $wrld
    }
    if {$world == ""} {
        set world [/socket:current]
    } elseif {[lsearch -exact [/socket:connectednames] $world] == -1} {
        set world [/socket:current]
    }
    $wname.world delete 0 end
    $wname.world insert end $world

    if {$delay > 0.0} {
        $wname.delay set $delay
    }

    if {$file != {}} {
        $wname.fileentry delete 0 end
        $wname.fileentry insert end $file
    }

    if {$prepend != {}} {
        $wname.prependentry delete 0 end
        $wname.prependentry insert end $prepend
    }

    if {$append != {}} {
        $wname.appendentry delete 0 end
        $wname.appendentry insert end $append
    }

    focus $wname.fileentry
    return
}


proc /sendfile_dlog {} {
    /send_dlog -wname .quotedlog \
            -world [/socket:foreground] \
            -file [/prefs:get last_sent_file] \
            -delay [/prefs:get last_sent_delay] \
            -prepend [/prefs:get last_sent_prepend] \
            -append [/prefs:get last_sent_append]
}

