global treb_world_type_map
set treb_world_type_map(tiny) "TinyMU*/MOO"
set treb_world_type_map(lp)   "LP/Diku/Aber"
set treb_world_type_map(lpp)  "LP/Diku w/ GOAHEAD/EOR"


proc /world {opt args} {
    dispatcher /world $opt $args
}


proc /world:edit {{item ""}} {
    if {$item == ""} {
        /editdlog Worlds World /world
    } else {
        /newdlog World /world $item
    }
}


package require opt

tcl::OptProc /world:add {
    {name     ""        "The name of the world."}
    {host     ""        "The host machine to connect to."}
    {port     -int 4201 "The port number to connect to."}
    {?char?   ""        "The character name."}
    {?passwd? ""        "The character password."}
    {-secure  -bool 0   "Encrypt connection with SSL or TLS."}
    {-noauth  -bool 0   "Allow unauthenticated encrypted connections."}
    {-script  ""        "The script to evaluate after connecting."}
    {-tcl     -bool 0   "Evaluate script as TCL if set to true."}
    {-type -choice {tiny lp lpp} "MUD server type.  lp & lpp allow unterminated prompt.  lp uses timeouts."}
} {
    global worlds
    set preexistant 0
    if {[info exists worlds($name)]} {
        set preexistant 1
    }
    set worlds($name) [list $host $port $char $passwd $tcl $script $secure $noauth $type]
    if {$preexistant} {
        /world:notifyregistered update $name
    } else {
        /world:notifyregistered add $name
    }
    global dirty_preferences; set dirty_preferences 1
    return "World added."
}


proc /world:delete {name} {
    global worlds
    if {![/world:exists $name]} {
        error "/world delete: No such World!"
    }
    unset worlds($name)
    /world:notifyregistered delete $name
    global dirty_preferences; set dirty_preferences 1
    return "World deleted."
}


proc /world:names {{pattern *}} {
    global worlds
    return [lsort -dictionary [array names worlds $pattern]]
}


proc /world:exists {name} {
    global worlds
    return [info exists worlds($name)]
}


proc /world:get {entry name} {
    global worlds
    if {![/world:exists $name]} {
        error "/world get: No such World!"
    }
    switch -exact -- $entry {
        host     {return [lindex $worlds($name) 0]}
        port     {return [lindex $worlds($name) 1]}
        charname {return [lindex $worlds($name) 2]}
        password {return [lindex $worlds($name) 3]}
        tcl      {return [lindex $worlds($name) 4]}
        script   {return [lindex $worlds($name) 5]}
        secure   {return [lindex $worlds($name) 6]}
        noauth   {return [lindex $worlds($name) 7]}
        type     {return [lindex $worlds($name) 8]}
        default  {error "/world get: Entry '$entry' must be one of host, port, secure, noauth, charname, password, tcl, script, noauth, or type."}
    }
}


proc /world:set {entry name value} {
    global worlds
    if {![/world:exists $name]} {
        error "/world set: No such World!"
    }
    set world $worlds($name)
    switch -exact -- $entry {
        host     {set worlds($name) [lreplace $world 0 0 $value]}
        port     {set worlds($name) [lreplace $world 1 1 $value]}
        charname {set worlds($name) [lreplace $world 2 2 $value]}
        password {set worlds($name) [lreplace $world 3 3 $value]}
        tcl      {set worlds($name) [lreplace $world 4 4 $value]}
        script   {set worlds($name) [lreplace $world 5 5 $value]}
        secure   {set worlds($name) [lreplace $world 6 6 $value]}
        noauth   {set worlds($name) [lreplace $world 7 7 $value]}
        type     {set worlds($name) [lreplace $world 8 8 $value]}
        default {error "/world set: Entry '$entry' must be one of host, port, secure, noauth, charname, password, tcl, script, noauth, or type."}
    }
    global dirty_preferences; set dirty_preferences 1
    return ""
}


proc /world:list {{pattern *}} {
    set oot ""
    foreach world [/world:names $pattern] {
        if {$oot != ""} {
            append oot "\n"
        }
        append oot "/world add [list $world]"
        append oot " [list [/world:get host $world]]"
        append oot " [list [/world:get port $world]]"
        append oot " [list [/world:get charname $world]]"
        append oot " [list [/world:get password $world]]"
        if {[/world:get tcl $world] != 0} { append oot " -tcl [list [/world:get tcl $world]]" }
        if {[/world:get script $world] != {}} { append oot " -script [list [/world:get script $world]]" }
        if {[/world:get secure $world] != 0} { append oot " -secure [list [/world:get secure $world]]" }
        if {[/world:get noauth $world] != 0} { append oot " -noauth [list [/world:get noauth $world]]" }
        if {[/world:get type $world] != "tiny"} { append oot " -type [list [/world:get type $world]]" }
    }
    return "$oot"
}


proc /world:connect {name} {
    global worlds
    if {![/world:exists $name]} {
        error "/world connect: No such World!"
    }
    /socket:open $name
    return ""
}


proc /world:notifyregistered {type name} {
    global WorldInfo
    foreach key [array names WorldInfo "reg,*"] {
        eval "$WorldInfo($key) [list $type] [list $name]"
    }
    return
}


proc /world:register {id cmd} {
    global WorldInfo
    if {[info exists WorldInfo(reg,$id)]} {
        error "/world register: That id is already registered!"
    }
    set WorldInfo(reg,$id) "$cmd"
    return
}


proc /world:deregister {id} {
    global WorldInfo
    if {[info exists WorldInfo(reg,$id)]} {
        unset WorldInfo(reg,$id)
    }
    return
}


proc /world:widgets {opt args} {
    dispatcher /world:widgets $opt $args
}


proc /world:widgets:create {master updatescript} {
    global treb_world_type_map
    set base $master.fr

    frame $base -relief flat -borderwidth 2
    label $base.namelbl -text {World Name} -anchor w
    label $base.hostlbl -text {Host Name} -anchor w
    label $base.portlbl -text {Port Number} -anchor w
    label $base.typelbl -text {Mud Type} -anchor w
    label $base.charlbl -text {Character} -anchor w
    label $base.passlbl -text {Password} -anchor w

    entry $base.name -textvariable WorldInfo($master,gui,name) -width 32
    entry $base.host -textvariable WorldInfo($master,gui,host) -width 32
    entry $base.port -textvariable WorldInfo($master,gui,port) -width 5
    checkbutton $base.secure -text "Encrypted" -command $updatescript \
        -variable WorldInfo($master,gui,secure)
    checkbutton $base.noauth -command $updatescript \
        -text "Allow unauthenticated encryption" \
        -variable WorldInfo($master,gui,noauth)
    combobox $base.type -textvariable WorldInfo($master,gui,type) \
        -changecommand "$updatescript" -editable 0
    foreach worldtype [array names treb_world_type_map] {
        $base.type entryinsert end $treb_world_type_map($worldtype)
    }
    $base.type delete 0 end
    $base.type insert end $treb_world_type_map(tiny)
    entry $base.char -textvariable WorldInfo($master,gui,char) -width 32
    entry $base.pass -textvariable WorldInfo($master,gui,pass) -show "*" -width 32
    bind $base.name <Key> +$updatescript
    bind $base.host <Key> +$updatescript
    bind $base.port <Key> +$updatescript
    bind $base.char <Key> +$updatescript
    bind $base.pass <Key> +$updatescript
    bind $base.name <<Cut>> +$updatescript
    bind $base.host <<Cut>> +$updatescript
    bind $base.port <<Cut>> +$updatescript
    bind $base.char <<Cut>> +$updatescript
    bind $base.pass <<Cut>> +$updatescript
    bind $base.name <<Paste>> +$updatescript
    bind $base.host <<Paste>> +$updatescript
    bind $base.port <<Paste>> +$updatescript
    bind $base.char <<Paste>> +$updatescript
    bind $base.pass <<Paste>> +$updatescript
    if {[info commands tls::init] == {}} {
        foreach cbctrl [list secure noauth] {
            $base.$cbctrl config -state disabled
            bind $base.$cbctrl <Button-1> "
                tk_messageBox -type ok -parent [winfo toplevel $master] \
                    -title {Trebuchet Notice} -icon warning \
                    -message {SSL support has not been installed.  Check http://www.belfry.com/fuzzball/trebuchet/ for info on installing SSL support.}
            "
        }
    }

    frame $base.script -relief flat -borderwidth 0
    label $base.script.lbl -text {Connect Script} -anchor w
    checkbutton $base.script.tclcb -text {Evaluate as TCL script} -onval 1 -offval 0 \
        -variable WorldInfo($master,gui,tcl) -command $updatescript
    text $base.script.script -width 40 -height 5 -yscrollcommand "$base.script.scroll set"
    scrollbar $base.script.scroll -orient vert -command "$base.script.script yview"
    bind $base.script.script <Key> +$updatescript
    bind $base.script.script <<Cut>> +$updatescript
    bind $base.script.script <<Paste>> +$updatescript

    grid columnconfig $base.script 0 -weight 1
    grid columnconfig $base.script 1 -minsize 10
    grid columnconfig $base.script 2 -minsize 10
    grid columnconfig $base.script 3 -minsize 10
    grid rowconfig $base.script 0 -minsize 10
    grid rowconfig $base.script 2 -weight 1
    grid $base.script.lbl    -row 1 -column 0 -sticky w
    grid $base.script.tclcb  -row 1 -column 2 -columnspan 2 -sticky e
    grid $base.script.script -row 2 -column 0 -columnspan 3 -sticky nsew
    grid $base.script.scroll -row 2 -column 3 -sticky ns

    grid columnconfig $base 2 -weight 1
    grid rowconfig $base 7 -weight 1

    grid $base.namelbl -row 0 -column 0 -sticky nse -padx 5 -pady 5
    grid $base.hostlbl -row 1 -column 0 -sticky nse -padx 5 -pady 5
    grid $base.portlbl -row 2 -column 0 -sticky nse -padx 5 -pady 5
    grid $base.typelbl -row 4 -column 0 -sticky nse -padx 5 -pady 5
    grid $base.charlbl -row 5 -column 0 -sticky nse -padx 5 -pady 5
    grid $base.passlbl -row 6 -column 0 -sticky nse -padx 5 -pady 5

    grid $base.name -row 0 -column 1 -columnspan 2 -sticky nsw -padx 10 -pady 5
    grid $base.host -row 1 -column 1 -columnspan 2 -sticky nsw -padx 10 -pady 5
    grid $base.port -row 2 -column 1 -sticky nsw -padx 10 -pady 5
    grid $base.secure -row 2 -column 2 -sticky nsw -padx 10 -pady 5
    grid $base.noauth -row 3 -column 1 -columnspan 2 -sticky nsw -padx 10 -pady 0
    grid $base.type -row 4 -column 1 -columnspan 2 -sticky nsw -padx 10 -pady 5
    grid $base.char -row 5 -column 1 -columnspan 2 -sticky nsw -padx 10 -pady 5
    grid $base.pass -row 6 -column 1 -columnspan 2 -sticky nsw -padx 10 -pady 5

    grid $base.script -row 8 -column 0 -columnspan 3 -sticky nsew -padx 10

    grid rowconf $master 1 -weight 1
    grid columnconf $master 1 -weight 1
    grid $base -pady 5 -row 0 -column 0 -sticky nsew

    return $base
}


proc /world:widgets:destroy {master} {
    destroy $master.fr
    return
}


proc /world:widgets:clear {master} {
    global WorldInfo
    set WorldInfo($master,gui,name)   ""
    set WorldInfo($master,gui,host)   ""
    set WorldInfo($master,gui,port)   ""
    set WorldInfo($master,gui,secure) ""
    set WorldInfo($master,gui,noauth) ""
    set WorldInfo($master,gui,char)   ""
    set WorldInfo($master,gui,pass)   ""
    set WorldInfo($master,gui,tcl)    ""
    set WorldInfo($master,gui,script) ""
    set WorldInfo($master,gui,type)   "TinyMUD/MUCK"
    return
}


proc /world:widgets:init {master name} {
    global WorldInfo
    global treb_world_type_map
    set type [/world get type $name]

    set WorldInfo($master,gui,name)   $name
    set WorldInfo($master,gui,host)   [/world get host $name]
    set WorldInfo($master,gui,port)   [/world get port $name]
    set WorldInfo($master,gui,secure) [/world get secure $name]
    set WorldInfo($master,gui,noauth) [/world get noauth $name]
    set WorldInfo($master,gui,char)   [/world get charname $name]
    set WorldInfo($master,gui,pass)   [/world get password $name]
    set WorldInfo($master,gui,tcl)    [/world get tcl $name]
    set WorldInfo($master,gui,type)   $treb_world_type_map($type)
    $master.fr.script.script delete 1.0 end
    $master.fr.script.script insert end [/world get script $name]
    return
}


proc /world:widgets:mknode {master} {
    global WorldInfo
    global treb_world_type_map

    foreach {key val} [array get treb_world_type_map] {
        set typemap($val) $key
    }
    set type $WorldInfo($master,gui,type)

    /world:add $WorldInfo($master,gui,name) \
        $WorldInfo($master,gui,host) \
        $WorldInfo($master,gui,port) \
        $WorldInfo($master,gui,char) \
        $WorldInfo($master,gui,pass) \
        -secure $WorldInfo($master,gui,secure) \
        -noauth $WorldInfo($master,gui,noauth) \
        -tcl $WorldInfo($master,gui,tcl) \
        -script [$master.fr.script.script get 1.0 end-1c] \
        -type $typemap($type)
    return
}


proc /world:widgets:getname {master} {
    global WorldInfo
    return $WorldInfo($master,gui,name)
}


proc /world:widgets:setname {master name} {
    global WorldInfo
    set WorldInfo($master,gui,name) $name
    return
}


proc /world:widgets:compare {master name} {
    global WorldInfo
    global treb_world_type_map
    set type [/world:get type $name]

    if {![/world:exists $name]} { return 0 }
    if {[/world:get host $name] != $WorldInfo($master,gui,host)} { return 0 }
    if {[/world:get port $name] != $WorldInfo($master,gui,port)} { return 0 }
    if {[/world:get charname $name] != $WorldInfo($master,gui,char)} { return 0 }
    if {[/world:get password $name] != $WorldInfo($master,gui,pass)} { return 0 }
    if {[/world:get secure $name] != $WorldInfo($master,gui,secure)} { return 0 }
    if {[/world:get noauth $name] != $WorldInfo($master,gui,noauth)} { return 0 }
    if {[/world:get tcl $name] != $WorldInfo($master,gui,tcl)} { return 0 }
    if {[/world:get script $name] != [$master.fr.script.script get 1.0 end-1c]} { return 0 }
    if {$treb_world_type_map($type) != $WorldInfo($master,gui,type)} { return 0 }
    return 1
}


