#
# TO DO:
#   Resetting game
#   Turn indicator
#   Undo move
#   Move only during your turn
#   Fix check and checkmate detection
#   Networked play via Trebuchet
#   En passant captures
#   Game handicapping
#   Allow user replacement graphics for pieces and tiles
#

load {} Tk

namespace eval chess {
	variable board
	variable pieces
	variable gamelist
	variable rgamelist
	variable piececount
	variable status_pid
	variable playercolor
	variable current_turn
	variable toplevcount 0
	variable sequencenum
	variable chess_package "org-fuzzball-chess"


	proc set_game_state {world base boardstate turn pcolor} {
		variable piececount
		variable current_turn
		variable playercolor

		set piecemap(P) "pawn"
		set piecemap(N) "knight"
		set piecemap(B) "bishop"
		set piecemap(R) "rook"
		set piecemap(Q) "queen"
		set piecemap(K) "king"

		set colormap(w) "white"
		set colormap(b) "black"

		remove_all_pieces $base
		for {set y 0} {$y < 8} {incr y} {
			for {set x 0} {$x < 8} {incr x} {
				set pos [expr {($y * 8) + $x}]
				set piece [string index $boardstate $pos]
				if {[string first $piece "PNBRQK"] != -1} {
					set color "W"
				} else {
					set color "B"
				}
				set pieceup [string toupper $piece]
				if {$pieceup != "." && $pieceup != " "} {
					set c [incr piececount($base,$color,$pieceup)]
					place_piece $world $base "$pieceup$c" $color $piecemap($pieceup) $x $y
				}
			}
		}
		set c [string tolower [string index $pcolor 0]]
		set t [string tolower [string index $turn 0]]
		set playercolor($base) $c
		set current_turn($base) $t
	}

	proc mkboard {base} {
		variable board

		set canvas $base.board
		# set blacktile black
		# set blacktile grey65
		# set blacktile #7faf9f
		set blacktile #9fcfbf
		set whitetile #DDDDDD
		for {set col 0} {$col < 8} {incr col} {
			$canvas create text [expr {$col * 32 + 16}] -8 -anchor center \
			    -justify center -text [string index "ABCDEFGH" $col]
			$canvas create text [expr {$col * 32 + 16}] 264 -anchor center \
			    -justify center -text [string index "ABCDEFGH" $col]
			$canvas create text  -9 [expr {$col * 32 + 16}] -anchor center \
			    -justify center -text [string index "87654321" $col]
			$canvas create text 263 [expr {$col * 32 + 16}] -anchor center \
			    -justify center -text [string index "87654321" $col]
			for {set row 0} {$row < 8} {incr row} {
				set square r${row}c${col}
				if {($row + $col) % 2 == 1} {
					set color $blacktile
				} else {
					set color $whitetile
				}
				set x [expr $col * 32]
				set y [expr $row * 32]
				set x2 [expr $x + 31]
				set y2 [expr $y + 31]
				$canvas create rectangle $x $y $x2 $y2 -fill $color -width 0 \
					-outline $color -tags $square
				set board($base,$col,$row) ""
			}
		}
		$canvas create rectangle -1 -1 [expr 8*32 - 1] [expr 8*32 - 1] \
		    -width 1 -outline black
	}

	proc init_pieces {} {
		foreach color {w b} {
			foreach type {pawn knight bishop rook queen king} {
				image create photo chess-$color$type -file images/$color$type.gif
			}
		}
	}

	proc place_piece {world base name color type col row} {
		variable pieces
		variable board

		set canvas $base.board
		set color [string tolower [string index $color 0]]
		set ref [string toupper "$color$name"]
		set img [string tolower "chess-$color$type"]
		set pieces($base,type,[string index $name 0]) $type
		set x [expr $col * 32]
		set y [expr $row * 32]
		$canvas create image $x $y -anchor nw -image $img -tags $ref
		$canvas bind $ref <ButtonPress-1>   "chess::dragstart $base $ref %x %y"
		$canvas bind $ref <B1-Motion>       "chess::dragpiece $base $ref %x %y"
		$canvas bind $ref <ButtonRelease-1> "chess::movepiece [list $world] $base $ref %x %y"
		set pieces($base,position,$ref) [list $col $row]
		set pieces($base,hasmoved,$ref) 0
		set board($base,$col,$row) $ref
	}

	proc setup_pieces {world base} {
		set stat "rnbqkbnrpppppppp................................PPPPPPPPRNBQKBNR"
		set_game_state $world $base $stat "White" "white"
	}

	proc remove_all_pieces {base} {
		variable pieces
		variable piececount

		foreach key [array names pieces $base,position,*] {
			set piece [string range $key [string length "$base,position,"] end]
			remove_piece $base $piece
		}
		foreach color {W B} {
			foreach piece {P N B R Q K} {
				set piececount($base,$color,$piece) 0
			}
		}
	}

	proc reset_game {world base} {
		remove_all_pieces $base
		setup_pieces $world $base
	}

	proc zip_back_timer {base item destx desty delx dely count donecmd} {
		set canvas $base.board
		$canvas move $item $delx $dely
		update idletasks
		incr count -1
		if {$count > 0} {
			after 10 "chess::zip_back_timer $base $item $destx $desty $delx $dely $count [list $donecmd]"
		} else {
			$canvas coords $item $destx $desty
			if {$donecmd != {}} {
				uplevel #0 $donecmd
			}
		}
	}

	proc zip_back {base item destx desty {donecmd {}}} {
		set canvas $base.board
		set pos [$canvas coords $item]
		set dx [expr $destx - [lindex $pos 0]]
		set dy [expr $desty - [lindex $pos 1]]
		set dist [expr sqrt(($dx * $dx) + ($dy * $dy)) ]
		set steps [expr 1 + int($dist / 10)]
		set delx [expr $dx / $steps]
		set dely [expr $dy / $steps]
		zip_back_timer $base $item $destx $desty $delx $dely $steps $donecmd
	}

	proc jumped_pieces {base startcol startrow endcol endrow} {
		variable board

		if {$startcol < $endcol} {
			set dcol 1
		} elseif {$startcol == $endcol} {
			set dcol 0
		} else {
			set dcol -1
		}
		if {$startrow < $endrow} {
			set drow 1
		} elseif {$startrow == $endrow} {
			set drow 0
		} else {
			set drow -1
		}
		set curcol $startcol
		set currow $startrow
		incr curcol $dcol
		incr currow $drow
		while {$curcol != $endcol || $currow != $endrow} {
			if {$board($base,$curcol,$currow) != ""} {
				return 1
			}
			incr curcol $dcol
			incr currow $drow
		}
		return 0
	}

	proc valid_capture {base item col row} {
		variable board
		variable pieces

		if {$row < 0 || $row > 7 || $col < 0 || $col > 7} {
			return 0
		}

		set mycolor [string range $item 0 0]
		set occupant $board($base,$col,$row)
		if {$occupant != ""} {
			set ocolor [string range $occupant 0 0]
			if {$ocolor == $mycolor} {
				return 0
			}
		}

		set curpos $pieces($base,position,$item)
		set curcol [lindex $curpos 0]
		set currow [lindex $curpos 1]
		set deltcol [expr $col - $curcol]
		set deltrow [expr $row - $currow]
		if {$deltrow == 0 && $deltcol == 0} {
			return 0
		}

		switch -regexp -- $item {
			BP[1-9] {
				if {$deltrow != 1 || abs($deltcol) != 1} {
					return 0
				}
			}
			WP[1-9] {
				if {$deltrow != -1 || abs($deltcol) != 1} {
					return 0
				}
			}
			default {
				return [valid_move $base $item $col $row]
			}
		}
		return 1
	}

	proc valid_castling {base item col row} {
		variable board
		variable pieces

		if {$row < 0 || $row > 7 || $col < 0 || $col > 7} {
			return 0
		}

		set curpos $pieces($base,position,$item)
		set curcol [lindex $curpos 0]
		set currow [lindex $curpos 1]
		set deltcol [expr $col - $curcol]
		set deltrow [expr $row - $currow]
		set occupant $board($base,$col,$row)

		if {[string toupper [string index $item 1]] != "K"} {
			return 0
		}
		if {!$pieces($base,hasmoved,$item) && $row == $currow} {
			if {[string index $occupant 1] == "R" &&
				!$pieces($base,hasmoved,$occupant) &&
				![jumped_pieces $base $curcol $currow $col $row]
			} {
				for {set tcol $curcol} {$tcol - $col != 0} {incr tcol $deltcol} {
					if {[in_check $base $item $tcol $row]} {
						return 0
					}
				}
				return 1
			}
		}
		return 0
	}

	proc valid_move {base item col row} {
		variable board
		variable pieces

		if {$row < 0 || $row > 7 || $col < 0 || $col > 7} {
			return 0
		}

		set mycolor [string range $item 0 0]
		set occupant $board($base,$col,$row)
		if {$occupant != ""} {
			set ocolor [string range $occupant 0 0]
			if {$ocolor == $mycolor} {
				return 0
			}
		}

		set curpos $pieces($base,position,$item)
		set curcol [lindex $curpos 0]
		set currow [lindex $curpos 1]
		set deltcol [expr $col - $curcol]
		set deltrow [expr $row - $currow]
		if {$deltrow == 0 && $deltcol == 0} {
			return 0
		}

		set do_check_check 1
		switch -regexp -- $item {
			BP[1-9] {
				if {$deltrow != 1} {
					if {$currow != 1 || $row != 3 || $deltcol != 0} {
						return 0
					}
					if {$board($base,$col,2) != ""} {
						return 0
					}
				}
				if {abs($deltcol) > 1} {
					return 0
				}
				if {$deltcol != 0 && $occupant == ""} {
					return 0
				}
				if {$deltcol == 0 && $occupant != ""} {
					return 0
				}
			}

			WP[1-9] {
				if {$deltrow != -1} {
					if {$currow != 6 || $row != 4 || $deltcol != 0} {
						return 0
					}
					if {$board($base,$col,5) != ""} {
						return 0
					}
				}
				if {abs($deltcol) > 1} {
					return 0
				}
				if {$deltcol != 0 && $occupant == ""} {
					return 0
				}
				if {$deltcol == 0 && $occupant != ""} {
					return 0
				}
			}

			[BW]R[1-9] {
				if {$deltcol != 0 && $deltrow != 0} {
					return 0
				}
				if {[jumped_pieces $base $curcol $currow $col $row]} {
					return 0
				}
			}

			[BW]N[1-9] {
				if {(abs($deltrow) != 2 || abs($deltcol) != 1) && 
					(abs($deltrow) != 1 || abs($deltcol) != 2)} {
					return 0
				}
			}

			[BW]B[1-9] {
				if {abs($deltcol) != abs($deltrow)} {
					return 0
				}
				if {[jumped_pieces $base $curcol $currow $col $row]} {
					return 0
				}
			}

			[BW]K1 {
				if {abs($deltcol) > 1 || abs($deltrow) > 1} {
					return 0
				}
				if {[in_check $base $item $col $row]} {
					return 0
				}
				set do_check_check 0
			}

			[BW]Q[1-9] {
				if {$deltcol == 0 || $deltrow == 0} {
					if {[jumped_pieces $base $curcol $currow $col $row]} {
						return 0
					}
				} else {
					if {abs($deltcol) != abs($deltrow)} {
						return 0
					}
					if {[jumped_pieces $base $curcol $currow $col $row]} {
						return 0
					}
				}
			}
		}
		if {$do_check_check} {
			set board($base,$col,$row) $item
			set board($base,$curcol,$currow) ""
			set myking "[string toupper $mycolor]K1"
			set kingpos $pieces($base,position,$myking)
			set kingcol [lindex $kingpos 0]
			set kingrow [lindex $kingpos 1]
			if {[in_check $base $myking $kingcol $kingrow $occupant]} {
				set board($base,$col,$row) $occupant
				set board($base,$curcol,$currow) $item
				return 0
			}
			set board($base,$col,$row) $occupant
			set board($base,$curcol,$currow) $item
		}
		return 1
	}

	proc in_check {base item col row {except ""}} {
		variable board
		variable pieces

		set color [string range $item 0 0]
		switch -exact -- $color {
			W {set ocolor B}
			B {set ocolor W}
			default {error "INTERNAL ERROR!"}
		}
		set curpos $pieces($base,position,$item)
		set curcol [lindex $curpos 0]
		set currow [lindex $curpos 1]
		set occupant $board($base,$col,$row)
		set board($base,$curcol,$currow) ""
		set board($base,$col,$row) $item
		foreach key [array names pieces $base,position,$ocolor*] {
			set piece [string range $key [string length "$base,position,"] end]
			if {$piece != $except && $piece != $occupant} {
				set result [valid_capture $base $piece $col $row]
				if {$result} {
					set board($base,$col,$row) $occupant
					set board($base,$curcol,$currow) $item
					return 1
				}
			}
		}
		set board($base,$col,$row) $occupant
		set board($base,$curcol,$currow) $item
		return 0
	}


	proc in_checkmate {base item col row} {
		if {![in_check $base $item $col $row]} {
			return 0
		}
		foreach {dcol drow} {-1 -1  -1 0  -1 1  0 -1  0 1  1 -1  1 0  1 1} {
			set icol [expr $col + $dcol]
			set irow [expr $row + $drow]
			if {[valid_move $base $item $icol $irow]} {
				if {![in_check $base $item $icol $irow]} {
					return 0
				}
			}
		}
		set color [string toupper [string index $item 0]]
		switch -exact -- $color {
			W {set ocolor B}
			B {set ocolor W}
			default {error "INTERNAL ERROR!"}
		}
		set threats {}
		variable pieces
		foreach key [array names pieces $base,position,$ocolor*] {
			set piece [string range $key [string length "$base,position,"] end]
			set result [valid_capture $base $piece $col $row]
			if {$result} {
				lappend threats $piece
			}
		}
		foreach threat $threats {
			set threatpos $pieces($base,position,$threat)
			set threatcol [lindex $threatpos 0]
			set threatrow [lindex $threatpos 1]
			foreach key [array names pieces $base,position,$color*] {
				set piece [string range $key [string length "$base,position,"] end]
				set result [valid_capture $base $piece $threatcol $threatrow]
				if {$result} {
					set piecepos $pieces($base,position,$piece)
					set piececol [lindex $threatpos 0]
					set piecerow [lindex $threatpos 1]
					set board($base,$piececol,$piecerow) ""
					set board($base,$threatcol,$threatrow) $piece
					set result [in_check $base $item $col $row $threat]
					set board($base,$piececol,$piecerow) $piece
					set board($base,$threatcol,$threatrow) $threat
					if {!$result} {
						return 0
					}
				}
			}
			if {[string match {[RBQ]} [string index $threat 1]]} {
				set dx [expr {$col - $threatcol}]
				set dy [expr {$col - $threatcol}]
				if {$dx != 0} {
					set dx [expr {$dx / abs($dx)}]
				}
				if {$dy != 0} {
					set dy [expr {$dy / abs($dy)}]
				}
				incr threatcol $dx
				incr threatrow $dy
				while {$threatcol != $col && $threatrow != $row} {
					foreach key [array names pieces $base,position,$color*] {
						set piece [string range $key [string length "$base,position,"] end]
						if {[string index $piece 1] == "K"} {
							continue
						}
						set result [valid_move $base $piece $threatcol $threatrow]
						if {$result} {
							set piecepos $pieces($base,position,$piece)
							set piececol [lindex $threatpos 0]
							set piecerow [lindex $threatpos 1]
							set board($base,$piececol,$piecerow) ""
							set board($base,$threatcol,$threatrow) $piece
							set result [in_check $base $item $col $row $threat]
							set board($base,$piececol,$piecerow) $piece
							set board($base,$threatcol,$threatrow) $threat
							if {!$result} {
								return 0
							}
						}
					}
					incr threatcol $dx
					incr threatrow $dy
				}
			}
		}
		return 1
	}


	proc encode_pos {col row} {
		set cchar [string range "abcdefgh" $col $col]
		set rchar [string range "87654321" $row $row]
		return "$cchar$rchar"
	}


	proc encode_move {base item startcol startrow destcol destrow occupant castling} {
		variable pieces

		set type [string index $item 1]
		set color [string index $item 0]
		switch -exact -- $color {
			B {set ocolor W}
			W {set ocolor B}
			default { error "INTERNAL ERROR" }
		}
		set oking ${ocolor}K1
		set okingpos $pieces($base,position,$oking)
		set okingcol [lindex $okingpos 0]
		set okingrow [lindex $okingpos 1]

		if {$castling} {
			if {$startcol > $destcol} {
				return "O-O-O  Queenside Castling."
			} else {
				return "O-O  Kingside Castling."
			}
		}

		set oot [encode_pos $startcol $startrow]
		if {$occupant != ""} {
			set otype [string index $occupant 1]
			set comment "  $pieces($base,type,$type) takes $pieces($base,type,$otype)."
			append oot "x"
		} else {
			set comment ""
			append oot "-"
		}
		append oot [encode_pos $destcol $destrow]

		if {[in_check $base $oking $okingcol $okingrow]} {
			append oot "+"
			if {$comment != ""} {
				append comment "  "
			}
			append comment "Check"
			if {[in_checkmate $base $oking $okingcol $okingrow]} {
				append oot "+"
				append comment "mate"
			}
			append comment "."
		}

		append oot "$comment"
		return $oot
	}

	proc decode_pos {pos} {
		set cnames "abcdefgh"
		set rnames "87654321"
		set pos [string tolower $pos]
		set col [string first [string index $pos 0] $cnames]
		set row [string first [string index $pos 1] $rnames]
		if {$col == -1 && $row == -1} {
			set col [string first [string index $pos 1] $cnames]
			set row [string first [string index $pos 0] $rnames]
		}
		return [list $col $row]
	}

	proc remove_piece {base item} {
		variable pieces
		variable board

		if {$item == ""} {
			return
		}
		set pos $pieces($base,position,$item)
		set col [lindex $pos 0]
		set row [lindex $pos 1]
		set board($base,$col,$row) ""
		unset pieces($base,position,$item)
		$base.board delete $item
	}

	proc capture_piece {base item} {
		remove_piece $base $item
	}

	proc promote_pawn {world base item} {
		variable pieces
		variable piececount

		if {$pieces($base,type,[string index $item 1]) != "pawn"} {
			return $item
		}
		foreach {col row} $pieces($base,position,$item) break;
		set color [string range $item 0 0]
		if {($color == "W" && $row == 0) || ($color == "B" && $row == 7)} {
			remove_piece $base $item
			set c [incr piececount($base,$color,Q)]
			set newpiece "Q$c"
			place_piece $world $base $newpiece $color "queen" $col $row
			set pieces($base,hasmoved,$newpiece) 1
			update idletasks
			return $color$newpiece
		}
		return $item
	}

	proc move_from {world base fromcol fromrow tocol torow} {
		variable board
		variable pieces
		variable current_turn
		variable playercolor

		set canvas $base.board
		set item $board($base,$fromcol,$fromrow)
		set oldcol $fromcol
		set oldrow $fromrow

		set col $tocol
		set row $torow
		set color [string tolower [string range $item 0 0]]
		set castlecol -1
		if {$color != $current_turn($base)} {
			setstatus $base 5 "Other player attempted to move out of turn!"
			return
		}
		set castling_is_valid [valid_castling $base $item $col $row]
		if {!$castling_is_valid && ![valid_move $base $item $col $row]} {
			setstatus $base 5 "Other player attempted invalid move!"
			return
		}
		set occupant $board($base,$col,$row)
		if {$castling_is_valid} {
			set castlecol $col
			set board($base,$col,$row) ""
			set board($base,$oldcol,$oldrow) ""
			if {$col > $oldcol} {
				set rookcol [expr $oldcol + 1]
				set newcol [expr $oldcol + 2]
			} else {
				set rookcol [expr $oldcol - 1]
				set newcol [expr $oldcol - 2]
			}
			set board($base,$rookcol,$row) $occupant
			set pieces($base,position,$occupant) [list $rookcol $row]
			set pieces($base,hasmoved,$occupant) 1
			set x [expr $rookcol * 32]
			set y [expr $row * 32]
			zip_back $base $occupant $x $y
			set occupant ""
			set col $newcol
		} else {
			set board($base,$oldcol,$oldrow) ""
			capture_piece $base $occupant
		}
		set pieces($base,hasmoved,$item) 1
		set board($base,$col,$row) $item
		set pieces($base,position,$item) [list $col $row]
		set x [expr $col * 32]
		set y [expr $row * 32]
		set item [promote_pawn $world $base $item]
		$canvas coords $item [expr {$oldcol * 32}] [expr {$oldrow * 32}]
		zip_back $base $item $x $y
		if {$castlecol >= 0} {
			dprint $base [encode_move $base $item $oldcol $oldrow $castlecol $row $occupant 1]
		} else {
			dprint $base [encode_move $base $item $oldcol $oldrow $col $row $occupant 0]
		}
		if {$current_turn($base) == "w"} {
			set current_turn($base) "b"
		} else {
			set current_turn($base) "w"
		}
	}

	proc movepiece {world base item x y} {
		variable board
		variable pieces
		variable current_turn
		variable playercolor
		variable rgamelist
		variable chess_package

		set canvas $base.board
		set oldpos $pieces($base,position,$item)
		set oldcol [lindex $oldpos 0]
		set oldrow [lindex $oldpos 1]

		set col [expr int(($x - 16) / 32)]
		set row [expr int(($y - 16) / 32)]
		set color [string tolower [string range $item 0 0]]
		set castlecol -1
		if {$color != $current_turn($base) || $color != $playercolor($base)} {
			zip_back $base $item [expr $oldcol * 32] [expr $oldrow * 32]
			setstatus $base 5 "Sorry, but it's not your turn!"
			return
		}
		set castling_is_valid [valid_castling $base $item $col $row]
		if {!$castling_is_valid && ![valid_move $base $item $col $row]} {
			zip_back $base $item [expr $oldcol * 32] [expr $oldrow * 32]
			return
		}
		set occupant $board($base,$col,$row)
		if {$castling_is_valid} {
			set castlecol $col
			set board($base,$col,$row) ""
			set board($base,$oldcol,$oldrow) ""
			if {$col > $oldcol} {
				set rookcol [expr $oldcol + 1]
				set newcol [expr $oldcol + 2]
			} else {
				set rookcol [expr $oldcol - 1]
				set newcol [expr $oldcol - 2]
			}
			set board($base,$rookcol,$row) $occupant
			set pieces($base,position,$occupant) [list $rookcol $row]
			set pieces($base,hasmoved,$occupant) 1
			set x [expr $rookcol * 32]
			set y [expr $row * 32]
			zip_back $base $occupant $x $y
			set occupant ""
			set col $newcol
		} else {
			set board($base,$oldcol,$oldrow) ""
			capture_piece $base $occupant
		}
		set pieces($base,hasmoved,$item) 1
		set board($base,$col,$row) $item
		set pieces($base,position,$item) [list $col $row]
		set x [expr $col * 32]
		set y [expr $row * 32]
		if {$castlecol >= 0} {
			# Castling.  No need to promote pawns.
			zip_back $base $item $x $y
			dprint $base [encode_move $base $item $oldcol $oldrow $castlecol $row $occupant 1]
		} else {
			$canvas coords $item $x $y
			promote_pawn $world $base $item
			dprint $base [encode_move $base $item $oldcol $oldrow $col $row $occupant 0]
		}

		set key $rgamelist($base)
		if {$castlecol >= 0} {
			set truecol $castlecol
		} else {
			set truecol $col
		}
		/mcp_send $chess_package "move" -world $world game $key \
				from [encode_pos $oldcol $oldrow] \
				to [encode_pos $truecol $row]

		if {$current_turn($base) == "w"} {
			set current_turn($base) "b"
			set c "black"
		} else {
			set current_turn($base) "w"
			set c "white"
		}
		$base.turn config -text "It's $c's move."
	}

	proc dragstart {base item x y} {
		variable current_turn
		variable playercolor

		set canvas $base.board
		$canvas raise $item
		set color [string tolower [string range $item 0 0]]
		if {$current_turn($base) != $color || $color != $playercolor($base)} {
			bell
			setstatus $base 5 "Sorry, but it's not your turn!"
			return
		}
		set x [expr $x - 36]
		set y [expr $y - 36]
		$canvas coords $item $x $y
	}

	proc dragpiece {base item x y} {
		variable current_turn
		variable playercolor

		set canvas $base.board
		$canvas raise $item
		set color [string tolower [string range $item 0 0]]
		if {$current_turn($base) != $color || $color != $playercolor($base)} {
			return
		}
		set x [expr $x - 36]
		set y [expr $y - 36]
		$canvas coords $item $x $y
	}

	proc clearstatus {base} {
		variable status_pid

		if {[winfo exists $base]} {
			set status_pid($base) ""
			$base.status config -text {}
		}
	}

	proc setstatus {base delay str} {
		variable status_pid

		$base.status config -text $str
		if {$status_pid($base) != ""} {
			after cancel $status_pid($base)
		}
		set status_pid($base) ""
		if {$delay > 0} {
			set status_pid($base) [after [expr int(1000 * $delay)] chess::clearstatus $base]
		}
	}

	proc dprint {base str} {
		$base.histfr.text insert end "$str\n"
		$base.histfr.text see end
	}

	proc free_board {base} {
		variable board

		for {set row 0} {$row < 8} {incr row} {
			for {set col 0} {$col < 8} {incr col} {
				unset board($base,$col,$row)
			}
		}
	}

	proc board:create {world base} {
        global treb_fonts
		toplevel $base
		wm title $base "Fuzzball Chess - TrebTk"
		wm protocol $base WM_DELETE_WINDOW "chess::GameEnd [list $world] $base"
		wm iconname $base "chess"
		wm group $base $base
		place_window_default $base

		menu $base.menu -tearoff 0
		$base.menu add command -label "Refresh" -underline 0 -command "chess::GameRefresh [list $world] $base"
		$base.menu add command -label "Close" -underline 0 -command "chess::GameEnd [list $world] $base"
		$base config -menu $base.menu
		label $base.black -text "Black: " -relief flat -anchor w -font [concat $treb_fonts(sansserif) bold]
		label $base.white -text "White: " -relief flat -anchor w -font [concat $treb_fonts(sansserif) bold]
		canvas $base.board -width [expr 9 * 32] -height [expr 9 * 32] \
			-scrollregion [list -16 -16 [expr 8*32 + 16 ] [expr 8*32 + 16]] \
			-relief solid -borderwidth 1
		label $base.turn -text "It is White's Turn" -relief flat \
		    -anchor w -font [concat $treb_fonts(sansserif) bold]
		label $base.status -text "Starting up" -relief flat -anchor w \
		    -wraplength 272 -justify left
		frame $base.histfr -relief sunken
		text $base.histfr.text -width 16 -height 5 -font $treb_fonts(fixed) \
		    -yscrollcommand "$base.histfr.scroll set"
		scrollbar $base.histfr.scroll -orient vert \
		    -command "$base.histfr.text yview"

		grid columnconfig $base 0 -weight 1
		grid columnconfig $base 1 -weight 1
		grid rowconfig $base 0 -weight 0
		grid rowconfig $base 1 -weight 0
		grid rowconfig $base 2 -weight 0
		grid rowconfig $base 3 -weight 0
		grid rowconfig $base 4 -weight 1

		grid $base.white  -row 0 -column 0 -columnspan 1 -sticky nw
		grid $base.black  -row 0 -column 1 -columnspan 1 -sticky nw
		grid $base.board  -row 1 -column 0 -columnspan 2 -sticky nw
		grid $base.turn   -row 2 -column 0 -columnspan 2 -sticky nsew
		grid $base.status -row 3 -column 0 -columnspan 2 -sticky nw
		grid $base.histfr -row 4 -column 0 -columnspan 2 -sticky nsew

		pack $base.histfr.scroll -side right -fill y -expand 0
		pack $base.histfr.text -side right -fill both -expand 1

		update
		mkboard $base
		setup_pieces $world $base
		update
		setstatus $base 0 ""
	}

	proc GameRefresh {world base} {
		variable rgamelist
		variable chess_package

		/mcp_send $chess_package refresh -world $world game $rgamelist($base)
	}

	proc GameEnd {world base} {
		variable piececount
		variable status_pid
		variable playercolor
		variable current_turn
		variable sequencenum
		variable gamelist
		variable rgamelist
		variable chess_package

		unset status_pid($base)
		unset playercolor($base)
		unset current_turn($base)
		unset sequencenum($base)

		set pkg $chess_package

		set key $rgamelist($base)
		/mcp_send $pkg leave -world $world game $key
		unset gamelist($key-$world)
		unset rgamelist($base)

		foreach color {W B} {
			foreach piece {P N B R Q K} {
				unset piececount($base,$color,$piece)
			}
		}
		remove_all_pieces $base
		free_board $base
		destroy $base
	}

	proc GameInit {world base} {
		variable piececount
		variable status_pid
		variable playercolor
		variable current_turn
		variable sequencenum

		set status_pid($base)   ""
		set playercolor($base)  "w"
		set current_turn($base) "w"
		set sequencenum($base)  0

		foreach color {W B} {
			foreach piece {P N B R Q K} {
				set piececount($base,$color,$piece) 0
			}
		}
		board:create $world $base
	}

	proc McpSetBoard {world data} {
		variable gamelist
		variable rgamelist
		variable toplevcount
		variable sequencenum
		variable chess_package

		set pkg $chess_package

		array set arr $data
		setfrom object   arr(object)   {}
		setfrom board    arr(board)    {}
		setfrom sequence arr(sequence)  0 int
		setfrom turn     arr(turn)     {White}
		setfrom color    arr(color)    {}
		setfrom bplayer  arr(blackplayer) {}
		setfrom wplayer  arr(whiteplayer) {}

		set rawobj $object
		append object "-$world"

		if {![info exists gamelist($object)]} {
		    incr toplevcount
			set gamelist($object) ".chess$toplevcount"
			set rgamelist($gamelist($object)) $rawobj
			GameInit $world $gamelist($object)
		}
		#if {$sequence <= $sequencenum($gamelist($object))} {
		#	return;
		#}
		$gamelist($object).white config -text "White: $wplayer"
		$gamelist($object).black config -text "Black: $bplayer"
		$gamelist($object).turn config -text "It's $turn's move."
		set_game_state $world $gamelist($object) $board $turn $color
		focus $gamelist($object)
		return
	}

	proc McpMove {world data} {
		variable gamelist
		variable toplevcount
		variable sequencenum
		variable chess_package

		set pkg $chess_package

		array set arr $data
		setfrom object   arr(game)     {}
		setfrom from     arr(from)     {}
		setfrom to       arr(to)       {}
		setfrom sequence arr(sequence)  0 int
		setfrom bplayer  arr(blackplayer) {}
		setfrom wplayer  arr(whiteplayer) {}
		setfrom turn     arr(turn)     {White}

		append object "-$world"

		if {![info exists gamelist($object)]} {
			return;
		}
        set base $gamelist($object)
		if {$sequence <= $sequencenum($base)} {
			return;
		}
		set sequencenum($base) $sequence
		set from [decode_pos $from]
		set to [decode_pos $to]
		$base.white config -text "White: $wplayer"
		$base.black config -text "Black: $bplayer"
		move_from $world $base \
				[lindex $from 0] [lindex $from 1] \
				[lindex $to 0] [lindex $to 1]
		$base.turn config -text "It is $turn's move."
        set current_turn($base) [string tolower [string index $turn 0]]
		return
	}

	proc McpInfo {world data} {
		variable gamelist
		variable toplevcount
		variable sequencenum
		variable chess_package

		set pkg $chess_package

		array set arr $data
		setfrom game arr(game) {}
		setfrom info arr(info) {}

		append game "-$world"

		if {![info exists gamelist($game)]} {
			return;
		}
		# dprint $gamelist($game) $info
		setstatus $gamelist($game) 5 $info
		return
	}

	proc Initialize {} {
		variable chess_package

		init_pieces
		mcp_register_pkg $chess_package 1.0 1.0
		mcp_register_handler $chess_package "setboard" chess::McpSetBoard
		mcp_register_handler $chess_package "move"     chess::McpMove
		mcp_register_handler $chess_package "info"     chess::McpInfo
	}
}

chess::Initialize

